from __future__ import absolute_import

import io
import ssl
import errno
import socket
import struct
import threading

from .. import logs
from .. import utils
from .. import errors
from .. import param

from . import Connection, Transport

BACKLOG = socket.SOMAXCONN
CHUNK_SIZE = io.DEFAULT_BUFFER_SIZE
ACCEPT_TIMEOUT = 0.1

log = logs.get(__name__)

class TcpConnection(Connection):
    log = log

    def __init__(self, interface, sock, url, chunk_size=None):
        super(TcpConnection, self).__init__(interface, url)
        self._sock = sock
        self._chunk_size = chunk_size
        self.log.debug('connected: %s', self.url)

    def recv(self):
        return recv(self._sock, self._chunk_size)

    def send(self, data):
        try:
            return send(self._sock, data)
        except socket.error as e:
            raise errors.TransportError(e)

    def close(self):
        close(self._sock)
        self.log.debug('disconnected: %s', self.url)

class TcpTransport(Transport):
    _name_ = 'tcp'
    log = log
    Connection = TcpConnection

    @param('backlog', int, default=BACKLOG)
    @param('chunk_size', int, default=CHUNK_SIZE)
    @param('ssl_key', doc='server-side only')
    def __init__(self, url, timeout=None, backlog=None, chunk_size=None,
            ssl_cert=None, ssl_key=None):
        super(TcpTransport, self).__init__(url)
        url = self._url

        self._addr = (url.host, url.port)
        self._sock = None

        self.timeout = timeout
        self.backlog = backlog or BACKLOG
        self.chunk_size = chunk_size

        self._ssl_cert = ssl_cert
        self._ssl_key = ssl_key

        self._stop = threading.Event()
        self._stopped = threading.Event()

    def connect(self, client):
        s = socket.create_connection(self._addr, self.timeout)

        if self._ssl_cert:
            ctx = ssl.create_default_context()
            ctx.load_verify_locations(self._ssl_cert)
            hostname = socket.gethostbyaddr(self._addr[0])[0]
            s = ctx.wrap_socket(s, server_hostname=hostname)

        return self.Connection(client, s, self._url.netloc, self.chunk_size)

    def bind(self):
        self._sock = s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(ACCEPT_TIMEOUT)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(self._addr)
        s.listen(self.backlog)

    def serve(self, server):
        self.bind()
        self.log.info('listening: %s', self.url)

        s = self._sock

        ctx = None
        if self._ssl_cert:
            ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ctx.load_cert_chain(certfile=self._ssl_cert, keyfile=self._ssl_key)

        stop = self._stop
        stopped = self._stopped
        stop.clear()
        stopped.clear()

        try:
            while not stop.is_set():
                try:
                    sock, addr = s.accept()
                except socket.timeout:
                    continue
                sock.settimeout(self.timeout)

                try:
                    if ctx:
                        sock = ctx.wrap_socket(sock, server_side=True)
                except ssl.SSLError:
                    log.exception('ssl error')
                    continue

                utils.start_thread(self.handle, server, sock, addr)
        finally:
            stopped.set()

    def handle(self, server, sock, addr):
        addr = utils.url.format_addr(addr)
        with self.Connection(server, sock, addr, self.chunk_size) as con:
            server.handle(con)

    def stop(self):
        self._stop.set()

    def join(self, timeout=None):
        self._stopped.wait(timeout)

##
## sockio
##

def close(sock):
    try:
        sock.shutdown(socket.SHUT_RDWR)
    except (OSError, socket.error) as e:
        # ignore if not connected
        if e.errno not in (errno.ENOTCONN,):# errno.EBADF):
            raise
    sock.close()

def recv(sock, chunk_size=None):
    """Receives a block of data from the socket."""
    try:
        return b''.join(recviter(sock, chunk_size))
    except errors.ReceiveInterrupted:
        return b''

def recviter(sock, chunk_size=None):
    """Returns an iterator over the data received from the socket.

    Each chunk of data yielded will be equal to or less than
    *chunk_size*.

    Can raise a `ReceiveInterrupted` exception.
    """
    buf = b''.join(recvsize(sock, 4, chunk_size))
    data_len = struct.unpack('>I', buf)[0]
    for chunk in recvsize(sock, data_len, chunk_size):
        yield chunk

def recvsize(sock, size, chunk_size=None):
    """Returns an iterator over the chunks received from the socket.

    If the socket stops receiving data before *size* bytes have been read, a
    `ReceiveInterrupted` exception will be raised.
    """
    pos = 0
    chunk_size = min(size, chunk_size or CHUNK_SIZE)
    while pos < size:
        chunk = sock.recv(min(size-pos, chunk_size))
        if not chunk:
            raise errors.ReceiveInterrupted()
        pos += len(chunk)
        yield chunk

def send(sock, data):
    """Sends *data* over the socket."""
    data_len = len(data)
    size = struct.pack('>I', data_len)
    sock.sendall(size)
    sock.sendall(data)



import snekrpc
from snekrpc import logs

class Service(snekrpc.Service):
    @snekrpc.command()
    def echo(self, value):
        return value

def main():
    logs.init()

    s = snekrpc.Server('tcp://localhost:1234')
    s.add_service(Service, alias='ex')
    s.serve()

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass



from __future__ import print_function

import json
import pprint
import datetime

from . import Formatter

class RawFormatter(Formatter):
    _name_ = 'raw'

    def print(self, res):
        print(self.format(res), end='')

class PprintFormatter(Formatter):
    _name_ = 'pprint'

    def print(self, res):
        pprint.pprint(res)

class JsonFormatter(Formatter):
    _name_ = 'json'

    def format(self, res):
        return json.dumps(res, default=self.encode)

    def encode(self, obj):
        if isinstance(obj, bytes):
            try:
                return obj.decode('utf8')
            except UnicodeDecodeError:
                return '<binary data>'
        elif isinstance(obj, datetime.datetime):
            return obj.isoformat()
        raise TypeError('unsupported type: %s' % type(obj))



import snekrpc
from snekrpc import logs

def main():
    logs.init()

    c = snekrpc.Client('tcp://localhost:1234')
    print(c.service_names())

    s = c.service('ex')
    print(s.echo('asdf'))

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass



import re
import copy
import inspect
import keyword
import tokenize
import collections

try:
    from inspect import signature, Parameter as Param
except ImportError:
    from funcsigs import signature, Parameter as Param

from .. import errors

from . import compat
from .encoding import to_unicode, to_bytes

# tokenize.Name should only be used as a fallback, since Py2 and Py3 define it
# differently
_rx_ident = re.compile(r'^{}$'.format(tokenize.Name))
def is_identifier(s):
    try:
        return s.isidentifier()
    except AttributeError:
        return bool(_rx_ident.match(s)) and not keyword.iskeyword(s)

def command(**hints):
    """Decorator that identifies a function to expose as an RPC command.

    *hints* can be used to set type hints for arguments accepted by the
    command. A type hint can be a type object or a string. If a type object is
    passed, it's `__name__` attribute will be used as the type hint. The type
    hint should indicate to a client developer the type of value that is
    expected. Use of standard Python types is recommended.

    To provide more parameter metadata, use the `@param()` decorator.
    """
    def decorator(func):
        cmd_meta = func.__dict__.setdefault('_meta', {})
        params = cmd_meta.setdefault('params', {})
        for name, hint in hints.items():
            hintstr = _hint_to_str(hint)

            if hintstr == 'stream':
                if 'stream' in cmd_meta:
                    err = 'only one stream param is possible: {}'
                    raise errors.ParameterError(err.format(name))
                cmd_meta['stream'] = name

            params.setdefault(name, {})['hint'] = hintstr
        return func
    return decorator

def param(name, hint=None, doc=None, hide=False, **metadata):
    """Decorator that assigns additional metadata, such as a doc string, to
    individual command parameters.

    The default *hint* is a unicode string: `str`.

    Where supported, *hide* can be used to prevent parameters from being
    displayed. This can be used to hide `**kwargs` when only certain parameters
    are expected.
    """
    def decorator(func):
        cmd_meta = func.__dict__.setdefault('_meta', {})
        params = cmd_meta.setdefault('params', {})
        param = params.setdefault(name, {})

        if hint:
            hintstr = _hint_to_str(hint)

            if hintstr == 'stream':
                if 'stream' in cmd_meta:
                    err = 'only one stream param is possible: {}'
                    raise errors.ParameterError(err.format(name))
                cmd_meta['stream'] = name

            param['hint'] = hintstr

        if doc: param['doc'] = doc
        if hide: param['hide'] = hide
        param.update(metadata)
        return func
    return decorator

def _hint_to_str(hint):
    """Internal. Converts type hints to strings."""
    if not hint:
        name = 'str'
    elif isinstance(hint, (compat.str, bytes)):
        name = hint
    else:
        name = hint.__name__
        if name == 'str':
            name = 'str' if compat.PY3 else 'bytes'
        elif name == 'unicode':
            name = 'str'
    return to_unicode(name)

def func_to_dict(func, remove_self=False):
    """Returns metadata for *func* in a dict.

    If *remove_self* is `True` ...
    """
    d = {
        'name': func.__name__,
        'doc': func.__doc__,
        }

    # we mutate these dicts, so make copies
    cmd_meta = copy.deepcopy(getattr(func, '_meta', {}))
    cmd_params = cmd_meta.pop('params', {})

    sig = signature(func)

    if remove_self:
        # remove 'self' parameter
        sig = sig.replace(parameters=list(sig.parameters.values())[1:])

    # add parameters
    params = d['params'] = []
    for p in sig.parameters.values():
        param = {
            'name': p.name,
            'kind': int(p.kind),
            }

        hint = param.get('hint')
        if p.default is not p.empty:
            param['default'] = p.default
            # if no hint, set an automatic hint based on the default value
            if not hint and p.default is not None:
                param['hint'] = type(p.default).__name__

        param.update(cmd_params.pop(p.name, {}))

        params.append(param)

    # add missing hints as params
    for name, param in cmd_params.items():
        param['name'] = name
        param.setdefault('kind', int(Param.KEYWORD_ONLY))
        params.append(param)

    # stream
    if 'stream' in cmd_meta:
        d['stream'] = cmd_meta.pop('stream')

    if inspect.isgeneratorfunction(func):
        d['isgen'] = True

    if cmd_meta:
        # add any left over metadata
        d['meta'] = cmd_meta

    return d

def dict_to_func(d, callback):
    """Returns a function based on the metadata dict *d*, as created by
    `func_to_dict`.

    *callback* must be a function which will be wrapped and called by the
    returned function.

    Note that `exec` is used to create the wrapping function. This exposes a
    potential security risk, however care is taken to ensure that the only user
    provided input that is executed consists of function parameter names which
    must be valid Python identifiers. It should therefore be safe to use from a
    security perspective.

    `POSITIONAL_ONLY` parameters are silently converted to
    `POSITIONAL_OR_KEYWORD` parameters.
    """
    sig_defn = []
    sig_call = []

    meta = copy.deepcopy(d)
    params = collections.OrderedDict()

    for p in meta.pop('params'):
        if not is_identifier(p['name']):
            raise ValueError('invalid parameter name: {}'.format(p['name']))
        kind = {
            0: Param.POSITIONAL_ONLY,
            1: Param.POSITIONAL_OR_KEYWORD,
            2: Param.VAR_POSITIONAL,
            3: Param.KEYWORD_ONLY,
            4: Param.VAR_KEYWORD,
            }[p['kind']]

        if kind == Param.KEYWORD_ONLY:
            # we assume these are part of kwargs
            continue
        elif kind == Param.VAR_POSITIONAL:
            name = '*' + p['name']
        elif kind == Param.VAR_KEYWORD:
            name = '**' + p['name']
        else:
            name = p['name']

        def_name = name
        if 'default' in p:
            def_name += '=' + repr(p['default'])

        sig_defn.append(def_name)
        sig_call.append(name)

        params[p.pop('name')] = p

    meta['params'] = params

    sig_defn = ', '.join(sig_defn)
    sig_call = ', '.join(sig_call)

    isgen = meta.pop('isgen', False)
    if isgen:
        tpl = 'def {}({}):\n  for x in callback({}): yield x'
    else:
        tpl = 'def {}({}): return callback({})'
    src = tpl.format(d['name'], sig_defn, sig_call)

    # compile
    g = {'callback': callback}
    exec(src, g) # pylint: disable=exec-used

    func = g[meta.pop('name')]
    func.__doc__ = meta.pop('doc', None)
    func.__dict__['_meta'] = meta

    return func

def get_func_name(name):
    # func.__name__ must be bytes in Python2
    return to_unicode(name) if compat.PY3 else to_bytes(name)



import socket
import threading
import subprocess

import pytest

import snekrpc
from snekrpc import utils

SERVICE_NAME = 'test'

PYTHON2 = utils.path.base_path('.venv2/bin/python')
PYTHON3 = utils.path.base_path('.venv3/bin/python')

TRANSPORT_URLS = [
    'tcp://127.0.0.1:7357',
    'unix:///tmp/snekrpc-test.sock',
    'http://127.0.0.1:7357',
    'http2://127.0.0.1:7357',
    ]

class Session(object):
    def __init__(self, interpreter, url, codec):
        self.interpreter = interpreter
        self.url = url
        self.codec = codec
        self.proc = None

    def get_client(self):
        return snekrpc.Client(self.url, retry_count=2)

    def start_server(self):
        cmd = [self.interpreter, '-m',
            'snekrpc', '-S', '-u', self.url, '-i', 'tests.service',
            '-s', 'test', '-s', 'file', '-vv']
        self.proc = subprocess.Popen(cmd)

    def stop_server(self):
        self.proc.terminate()

def start_thread(func, *args, **kwargs):
    t = threading.Thread(target=func, args=args, kwargs=kwargs)
    t.daemon = True
    t.start()
    return t

@pytest.fixture(scope='session', autouse=True)
def socket_timeout():
    socket.setdefaulttimeout(5)

@pytest.fixture(scope='session', params=TRANSPORT_URLS)
def url(request):
    return request.param

@pytest.fixture(scope='session', params=['json', 'msgpack'])
def codec(request):
    return request.param

@pytest.fixture(scope='session', params=[PYTHON2, PYTHON3])
def session(request, url, codec):
    s = Session(request.param, url, codec)
    s.start_server()
    yield s
    s.stop_server()

@pytest.fixture
def service(session):
    return session.get_client().service(SERVICE_NAME)



import inspect
import traceback

from .op import Op
from . import logs
from . import utils
from . import errors

log = logs.get(__name__)

class Protocol(object):
    def __init__(self, interface, con, metadata=None):
        self._ifc = interface
        self._con = con

        self.metadata = metadata or {}

    @property
    def local_url(self):
        return self._ifc.url

    @property
    def remote_url(self):
        return self._con.url

    ## protocol ##

    def handle(self):
        recv = self._con.recv_msg

        while True:
            try:
                msg = recv()

                if msg is None:
                    return
                elif msg.op == Op.command:
                    self.recv_cmd(msg)
                else:
                    raise errors.ProtocolOpError(msg.op)

            except errors.TransportError as e:
                logger = (log.exception
                    if log.isEnabledFor(logs.DEBUG) else log.error)
                logger('transport error (%s): %s',
                    self.remote_url, utils.format.format_exc(e))

            except Exception as e:
                self.send_err(e)

    def recv_cmd(self, msg):
        """Processes commands received from the remote interface."""
        svc_name, cmd_name, args, kwargs = msg.data

        svc = self._ifc.service(svc_name)
        func = getattr(svc, cmd_name)

        # convert any stream to a generator
        recv_args = []
        recv_kwargs = {}

        for arg in args:
            if inspect.isgenerator(arg):
                arg = self.recv_stream()
            recv_args.append(arg)
        for name, arg in kwargs.items():
            if inspect.isgenerator(arg):
                arg = self.recv_stream()
            recv_kwargs[name] = arg

        args = recv_args
        kwargs = recv_kwargs

        if log.isEnabledFor(logs.DEBUG):
            log.debug('cmd: %s <- %s',
                utils.format.format_cmd(svc_name, cmd_name, args, kwargs),
                self.remote_url)

        # call
        res = func(*args, **kwargs)

        if inspect.isgenerator(res):
            self.send_stream(res)
        else:
            self._con.send_msg(Op.data, res)

    def send_cmd(self, svc_name, cmd_name, *args, **kwargs):
        """Sends a command operation to the remote interface."""
        if log.isEnabledFor(logs.DEBUG):
            log.debug('cmd: %s -> %s',
                utils.format.format_cmd(svc_name, cmd_name, args, kwargs),
                self.remote_url)

        # convert any generator arg to a stream
        stream = None
        send_args = []
        send_kwargs = {}

        for arg in args:
            if inspect.isgenerator(arg):
                if stream is not None:
                    err = 'only one stream param is possible'
                    raise errors.ParameterError(err)
                stream = arg
            send_args.append(arg)
        for name, arg in kwargs.items():
            if inspect.isgenerator(arg):
                if stream is not None:
                    err = 'only one stream param is possible'
                    raise errors.ParameterError(err)
                stream = arg
            send_kwargs[name] = arg

        args = send_args
        kwargs = send_kwargs

        # send command
        self._con.send_msg(Op.command, (svc_name, cmd_name, args, kwargs))

        # send stream
        if stream:
            self.send_stream(stream)

        msg = self._con.recv_msg()

        if msg.op == Op.data:
            return msg.data
        elif msg.op == Op.error:
            raise errors.RemoteError(*msg.data)
        elif msg.op == Op.stream_start:
            # have to return a generator (see service/__init__.py:wrap_call)
            return self.recv_stream(started=True)
        else:
            raise errors.ProtocolOpError(msg.op)

    def recv_stream(self, started=False):
        recv = self._con.recv_msg

        if not started:
            msg = recv()
            if not msg:
                raise errors.ReceiveInterrupted()
            if msg.op != Op.stream_start:
                raise errors.ProtocolOpError(msg.op)

        while True:
            msg = recv()
            if not msg:
                raise errors.ReceiveInterrupted()
            elif msg.op == Op.data:
                yield msg.data
            elif msg.op == Op.error:
                raise errors.RemoteError(*msg.data)
            elif msg.op == Op.stream_end:
                return
            else:
                raise errors.ProtocolOpError(msg.op)

    def send_stream(self, it):
        send = self._con.send_msg
        send(Op.stream_start)
        for x in it:
            send(Op.data, x)
        send(Op.stream_end)

    def send_err(self, e):
        name = e.__class__.__name__
        msg = str(e)
        tb = traceback.format_exc().rstrip() if self._ifc.remote_tracebacks else ''

        log.exception('%s: %s', name, msg)
        self._con.send_msg(Op.error, utils.encoding.to_unicode((name, msg, tb)))



from .. import Service, command, param
from ..utils.encoding import to_unicode
from . import service_to_dict as s2d

class MetadataService(Service):
    """Provides commands that return server and service metadata.

    Note that this service is handled as a special case in several places.
    """
    _name_ = 'meta'

    @param('server', hide=True)
    def __init__(self, server):
        self._server = server

    @command()
    def status(self):
        ifc = self._server
        res = {
            'codec': ifc.codec._name_,
            'transport': ifc.transport._name_,
            'version': ifc.version,
            }
        return to_unicode(res)

    @command()
    def service_names(self):
        return self._server.service_names()

    @command()
    def services(self):
        return to_unicode([s2d(svc) for svc in self._server.services()])

    @command()
    def service(self, name):
        return to_unicode(s2d(self._server.service(name)))



import os
import pkgutil
import zipfile
import argparse

from . import logs

SHELL_HEADER = b'#! /usr/bin/env python\n'
MAIN = b'from snekrpc import __main__;__main__.main()'
EXCLUDE = frozenset(['__pycache__'])

PACKAGES = frozenset(['msgpack', 'temporenc'])

log = logs.get(__name__)

def exclude(name):
    if name.startswith('.'):
        return True
    elif name.endswith(('.pyc', '.so')):
        return True
    elif name in EXCLUDE:
        return True
    return False

def collect_source(root_path):
    dst_root_path = os.path.dirname(root_path)

    for dpath, dnames, fnames in os.walk(root_path):
        dnames[:] = [dname for dname in dnames if not exclude(dname)]

        for fname in fnames:
            if exclude(fname):
                continue

            src_path = os.path.join(dpath, fname)
            rpath = os.path.relpath(dpath, dst_root_path)
            dst_path = os.path.join(rpath, fname)

            yield src_path, dst_path

def pack(root_path, output_path, services, compresslevel):
    paths = list(collect_source(root_path))

    for package in PACKAGES:
        modpath = pkgutil.get_loader(package).get_filename()
        paths += list(collect_source(os.path.dirname(modpath)))

    for service in services:
        modpath = pkgutil.get_loader(service).get_filename()
        modname = os.path.basename(modpath)
        paths.append((modpath, os.path.join('snekrpc/service', modname)))

    with open(output_path, 'wb') as fp:
        fp.write(SHELL_HEADER)

        log.debug('compression level set to: %s', compresslevel)
        z = zipfile.ZipFile(fp, 'a', zipfile.ZIP_DEFLATED,
            compresslevel=compresslevel)
        with z as zfp:
            fpath = '__main__.py'
            log.debug('adding: %s', fpath)
            zfp.writestr(fpath, MAIN)

            for src_path, dst_path in sorted(paths):
                log.debug('adding: %s', dst_path)
                zfp.write(src_path, dst_path)

    os.chmod(output_path, 0o744)
    log.info('packed script written to: %s', output_path)

def add_arguments(parser):
    parser.add_argument('output', help='path for the output file')
    parser.add_argument('-s', '--service',
        action='append', dest='services', default=[],
        help='a service module to include (can be set multiple times)')
    parser.add_argument('-c', '--compression-level', type=int, default=0,
        help='level of compression to apply (0-9) (default: %(default)s)')
    parser.add_argument('-v', '--verbose', action='count',
        default=0, help='enable verbose output')

def main():
    parser = argparse.ArgumentParser()
    add_arguments(parser)

    args = parser.parse_args()

    logs.init(args.verbose)

    pack('./snekrpc', args.output, args.services, args.compression_level)

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass



import threading

from .. import logs

# imports for convenience
from . import url
from . import path
from . import retry
from . import compat
from . import format
from . import encoding
from . import function

DEFAULT_URL = 'tcp://127.0.0.1:12321'

log = logs.get(__name__)

##
## threading
##

def start_thread(func, *args, **kwargs):
    def safe(func, *args, **kwargs):
        tid = threading.current_thread().ident
        log.debug('thread started [%s]: %s', tid, func.__name__)
        try:
            return func(*args, **kwargs)
        except Exception:
            log.exception('thread error')
        finally:
            log.debug('thread stopped [%s]: %s', tid, func.__name__)

    t = threading.Thread(target=safe, args=(func,) + args, kwargs=kwargs)
    t.daemon = True
    t.start()
    return t



import struct

from ..op import Op
from .. import logs
from .. import utils
from .. import errors
from .. import protocol
from .. import registry

TransportMeta = registry.create_metaclass(__name__)

log = logs.get(__name__)

def get(url, transport_args=None):
    """Returns an instance of the Transport matching the schema of *url*."""
    if isinstance(url, Transport):
        return url

    scheme = utils.url.Url(url).scheme
    cls = TransportMeta.get(scheme)
    if isinstance(cls, Exception):
        raise cls
    return cls(url, **(transport_args or {}))

class Transport(utils.compat.with_metaclass(TransportMeta, object)):
    def __init__(self, url):
        self._url = utils.url.Url(url)

    @property
    def url(self):
        return self._url

    def connect(self, client):
        raise NotImplementedError('abstract')

    def serve(self, server):
        raise NotImplementedError('abstract')

    def stop(self):
        raise NotImplementedError('abstract')

    def join(self, timeout=None):
        raise NotImplementedError('abstract')

class Message(object):
    def __init__(self, op, data=None):
        self.op = op
        self.data = data

    def __repr__(self):
        tpl = '{}(op=<{}:{}>{})'
        class_name = self.__class__.__name__
        data_str = ', data={}'.format(
            utils.format.elide(repr(self.data))) if self.data is not None else ''
        return tpl.format(class_name, self.op, Op.to_str(self.op), data_str)

class Connection(object):
    def __init__(self, interface, addr):
        self._ifc = interface
        self._addr = addr
        self._proto = protocol.Protocol(interface, self)

    @property
    def url(self):
        return self._addr

    def get_protocol(self, metadata=None):
        if metadata:
            self._proto.metadata = metadata
        return self._proto

    ## data protocol ##

    def send(self, data):
        raise NotImplementedError('abstract')

    def recv(self):
        raise NotImplementedError('abstract')

    ## handshake protocol ##

    def req_handshake(self):
        if self._ifc.codec:
            return

        if log.isEnabledFor(logs.DEBUG):
            log.debug('msg: %s -> %s', Message(Op.handshake, None), self._addr)

        buf = struct.pack('>B', Op.handshake)
        self.send(buf)

        buf = self.recv()
        op, codec = struct.unpack('>B{}s'.format(len(buf) - 1), buf)
        if op != Op.handshake:
            raise errors.ProtocolOpError(op)

        if log.isEnabledFor(logs.DEBUG):
            log.debug('msg: %s <- %s', Message(op, codec), self._addr)

        self._ifc.codec = codec.decode('utf8')

    def res_handshake(self, data):
        if data != b'\x00': # bit of a shortcut. skip conversion to Op.handshake
            return data

        if log.isEnabledFor(logs.DEBUG):
            log.debug('msg: %s <- %s', Message(Op.handshake, None), self._addr)

        name = self._ifc.codec._name_.encode('utf8')

        if log.isEnabledFor(logs.DEBUG):
            log.debug('msg: %s -> %s', Message(Op.handshake, name), self._addr)

        buf = struct.pack('>B{}s'.format(len(name)), Op.handshake, name)
        self.send(buf)

        # continue with standard protocol
        return self.recv()

    ## message protocol ##

    def recv_msg(self):
        data = self.recv()
        # check for handshake
        data = self.res_handshake(data)

        if not data:
            return None

        msg = Message(*self._ifc.codec._decode(data))

        if log.isEnabledFor(logs.DEBUG):
            log.debug('msg: %s <- %s', msg, self._addr)

        return msg

    def send_msg(self, op, data=None):
        # check for handshake
        self.req_handshake()

        if log.isEnabledFor(logs.DEBUG):
            log.debug('msg: %s -> %s', Message(op, data), self._addr)

        msg = self._ifc.codec._encode((op, data))
        return self.send(msg)

    ## management ##

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, etype, evalue, tb):
        self.close()



# encoding: utf-8

import sys
import inspect
import contextlib

import pytest

from snekrpc.utils import compat
from snekrpc.utils import function
from funcsigs import Parameter as Param

@contextlib.contextmanager
def does_not_raise():
    yield

class F(object):
    def null(self):
        pass

    def positional(self, a):
        'positional'
        return a

    def default(self, a=None):
        'default'
        return a

    @function.command(a=bool)
    def command(self, a):
        'command'
        return a

    def stream(self):
        'stream'
        for i in range(5):
            yield i

    # SyntaxError on Python < 3.8
    try:
        exec("def positional_only(self, a, /): 'positional_only'; return a")
    except SyntaxError:
        pass

    def var_positional(self, *a):
        'var_positional'
        return a

    def var_keyword(self, **a):
        'var_keyword'
        return a

    def mixed_params(self, a, b=None, *c, **d):
        'mixed_params'
        return a, b, c, d

    def default_no_hint(self, a=42):
        'default_no_hint'
        return a

    @function.param('a', int)
    def kwargs_param(self, **kwargs):
        'kwargs_param'
        return kwargs

    @function.param('a', compat.str, 'a param', extra=42)
    def param_decorator(self, a):
        'param_decorator'
        return a

def test_roundtrip():
    f1 = F().mixed_params
    d1 = function.func_to_dict(f1)
    f2 = function.dict_to_func(d1, f1)
    d2 = function.func_to_dict(f2)

    assert d1 == d2
    assert f1(1, 2, 3, 4, d=5, e=6) == f2(1, 2, 3, 4, d=5, e=6)

def test_roundtrip_generator():
    f1 = F().stream
    d1 = function.func_to_dict(f1)
    f2 = function.dict_to_func(d1, f1)
    d2 = function.func_to_dict(f2)

    assert d1 == d2
    assert inspect.isgeneratorfunction(f1)
    assert inspect.isgeneratorfunction(f2)
    assert list(f1()) == list(f2())

##
## dict_to_func tests
##

def test_d2f_null():
    f = F().null
    d = {
        'name': 'null',
        'params': [],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    assert func() == f()

def test_d2f_positional():
    f = F().positional
    d = {
        'name': 'positional',
        'doc': 'positional',
        'params': [{'name': 'a', 'kind': 1}],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    assert func(1) == f(1)

def test_d2f_default():
    f = F().default
    d = {
        'name': 'default',
        'doc': 'default',
        'params': [{'name': 'a', 'kind': 1, 'default': None}],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    assert func() == f()
    assert func(1) == f(1)

def test_d2f_command():
    f = F().command
    d = {
        'name': 'command',
        'doc': 'command',
        'params': [{'name': 'a', 'hint': 'bool', 'kind': 1}],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    assert func(True) == f(True)

def test_d2f_stream():
    f = F().stream
    d = {
        'name': 'stream',
        'doc': 'stream',
        'isgen': True,
        'params': [],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__
    assert inspect.isgeneratorfunction(func)

    assert list(func()) == list(f())

@pytest.mark.skipif(sys.version_info < (3, 8), reason='requires Python >= 3.8')
def test_d2f_positional_only():
    f = F().positional_only
    d = {
        'name': 'positional_only',
        'doc': 'positional_only',
        'params': [{'name': 'a', 'kind': 0}],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    assert func(1) == f(1)

    # expected: created function does not honor POSITIONAL_ONLY
    assert func(a=1) == f(1)

def test_d2f_var_positional():
    f = F().var_positional
    d = {
        'name': 'var_positional',
        'doc': 'var_positional',
        'params': [{'name': 'a', 'kind': 2}],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    for i in range(5):
        args = list(range(i))
        assert func(*args) == f(*args)

def test_d2f_var_keyword():
    f = F().var_keyword
    d = {
        'name': 'var_keyword',
        'doc': 'var_keyword',
        'params': [{'name': 'a', 'kind': 4}],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    for i in range(5):
        args = {'arg{}'.format(j): j for j in range(i)}
        assert func(**args) == f(**args)

def test_d2f_mixed_params():
    f = F().mixed_params
    d = {
        'name': 'mixed_params',
        'doc': 'mixed_params',
        'params': [
            {'kind': 1, 'name': 'a'},
            {'kind': 1, 'name': 'b', 'default': None},
            {'kind': 2, 'name': 'c'},
            {'kind': 4, 'name': 'd'},
            ],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    assert func(1, 2, 3, 4, d=5, e=6) == f(1, 2, 3, 4, d=5, e=6)

def test_d2f_default_no_hint():
    f = F().default_no_hint
    d = {
        'name': 'default_no_hint',
        'doc': 'default_no_hint',
        'params': [{'name': 'a', 'kind': 1, 'hint': 'int', 'default': 42}],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    assert func() == f() == 42
    assert func(666) == f(666) == 666

def test_d2f_kwargs_param():
    f = F().kwargs_param
    d = {
        'name': 'kwargs_param',
        'doc': 'kwargs_param',
        'params': [
            {'name': 'kwargs', 'kind': 4},
            {'name': 'a', 'kind': 3, 'hint': 'int'},
            ],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    with pytest.raises(TypeError):
        assert func(1) == f(1)
    assert func() == f()
    assert func(a=1) == f(a=1)
    assert func(a=1, x=2) == f(a=1, x=2)
    assert func(x=2) == f(x=2)

def test_d2f_param_decorator():
    f = F().param_decorator
    d = {
        'name': 'param_decorator',
        'doc': 'param_decorator',
        'params': [{
            'name': 'a',
            'kind': 1,
            'hint': 'str',
            'doc': 'a param',
            'extra': 42,
            }],
        }
    func = function.dict_to_func(d, f)

    assert func.__name__ == f.__name__
    assert func.__doc__ is f.__doc__

    assert func(1) == f(1)

@pytest.mark.parametrize('name, expectation', [
    ('sys.path', pytest.raises(ValueError)),
    ('l[0]', pytest.raises(ValueError)),
    ('l()', pytest.raises(ValueError)),
    ('1name', pytest.raises(ValueError)),
    ('#name', pytest.raises(ValueError)),
    ('__import__', does_not_raise()),
    ('name1', does_not_raise()),
    ])
def test_d2f_invalid_param_name(name, expectation):
    d = {
        'name': 'invalid',
        'params': [{'name': name, 'kind': 1}]
        }
    with expectation:
        function.dict_to_func(d, lambda: None)

##
## func_to_dict tests
##

def test_f2d_null():
    d = function.func_to_dict(F().null)
    assert d == {
        'name': 'null',
        'doc': None,
        'params': [],
        }

def test_f2d_positional():
    d = function.func_to_dict(F().positional)
    assert d == {
        'name': 'positional',
        'doc': 'positional',
        'params': [{'name': 'a', 'kind': 1}],
        }
    assert d['params'][0]['kind'] == Param.POSITIONAL_OR_KEYWORD

def test_f2d_default():
    d = function.func_to_dict(F().default)
    assert d == {
        'name': 'default',
        'doc': 'default',
        'params': [{'name': 'a', 'kind': 1, 'default': None}],
        }
    assert d['params'][0]['kind'] == Param.POSITIONAL_OR_KEYWORD

def test_f2d_command():
    d = function.func_to_dict(F().command)
    assert d == {
        'name': 'command',
        'doc': 'command',
        'params': [{'name': 'a', 'hint': 'bool', 'kind': 1}],
        }
    assert d['params'][0]['kind'] == Param.POSITIONAL_OR_KEYWORD

def test_f2d_stream():
    d = function.func_to_dict(F().stream)
    assert d == {
        'name': 'stream',
        'doc': 'stream',
        'isgen': True,
        'params': [],
        }

@pytest.mark.skipif(sys.version_info < (3, 8), reason='requires Python >= 3.8')
def test_f2d_positional_only():
    d = function.func_to_dict(F().positional_only)
    assert d == {
        'name': 'positional_only',
        'doc': 'positional_only',
        'params': [{'name': 'a', 'kind': 0}],
        }
    assert d['params'][0]['kind'] == Param.POSITIONAL_ONLY

def test_f2d_var_positional():
    d = function.func_to_dict(F().var_positional)
    assert d == {
        'name': 'var_positional',
        'doc': 'var_positional',
        'params': [{'name': 'a', 'kind': 2}],
        }
    assert d['params'][0]['kind'] == Param.VAR_POSITIONAL

def test_f2d_var_keyword():
    d = function.func_to_dict(F().var_keyword)
    assert d == {
        'name': 'var_keyword',
        'doc': 'var_keyword',
        'params': [{'name': 'a', 'kind': 4}],
        }
    assert d['params'][0]['kind'] == Param.VAR_KEYWORD

def test_f2d_mixed_params():
    d = function.func_to_dict(F().mixed_params)
    assert d == {
        'name': 'mixed_params',
        'doc': 'mixed_params',
        'params': [
            {'kind': 1, 'name': 'a'},
            {'kind': 1, 'name': 'b', 'default': None},
            {'kind': 2, 'name': 'c'},
            {'kind': 4, 'name': 'd'},
            ],
        }
    assert d['params'][0]['kind'] == Param.POSITIONAL_OR_KEYWORD
    assert d['params'][1]['kind'] == Param.POSITIONAL_OR_KEYWORD
    assert d['params'][2]['kind'] == Param.VAR_POSITIONAL
    assert d['params'][3]['kind'] == Param.VAR_KEYWORD

def test_f2d_default_no_hint():
    d = function.func_to_dict(F().default_no_hint)
    assert d == {
        'name': 'default_no_hint',
        'doc': 'default_no_hint',
        'params': [{'name': 'a', 'kind': 1, 'hint': 'int', 'default': 42}],
        }
    assert d['params'][0]['kind'] == Param.POSITIONAL_OR_KEYWORD

def test_f2d_kwargs_param():
    d = function.func_to_dict(F().kwargs_param)
    assert d == {
        'name': 'kwargs_param',
        'doc': 'kwargs_param',
        'params': [
            {'name': 'kwargs', 'kind': 4},
            {'name': 'a', 'kind': 3, 'hint': 'int'},
            ],
        }
    assert d['params'][0]['kind'] == Param.VAR_KEYWORD
    assert d['params'][1]['kind'] == Param.KEYWORD_ONLY

def test_f2d_param_decorator():
    d = function.func_to_dict(F().param_decorator)
    assert d == {
        'name': 'param_decorator',
        'doc': 'param_decorator',
        'params': [{
            'name': 'a',
            'kind': 1,
            'hint': 'str',
            'doc': 'a param',
            'extra': 42,
            }],
        }
    assert d['params'][0]['kind'] == Param.POSITIONAL_OR_KEYWORD



import posixpath

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

DEFAULT_SCHEME = 'tcp'
DEFAULT_HOST = '127.0.0.1'
DEFAULT_PORT = 12321

def format_addr(addr):
    return '{}:{}'.format(*addr)

class Url(object):
    def __init__(self, url):
        if isinstance(url, Url):
            self.__dict__ = url.__dict__
            return

        if ':/' not in url:
            url = '{}://{}'.format(DEFAULT_SCHEME, url)

        u = urlparse(url)

        self._scheme = u.scheme

        if self._scheme == 'unix':
            self._host = None
            self._port = None
            self._path = posixpath.join(
                u.hostname or posixpath.sep,
                u.path.lstrip(posixpath.sep)).rstrip(posixpath.sep)
            self._address = self._netloc = self._path
        else:
            if u.path.strip(posixpath.sep):
                raise ValueError('invalid URL: {}'.format(url))
            self._host = (u.hostname or DEFAULT_HOST).replace('*', '0.0.0.0')
            self._port = u.port or DEFAULT_PORT
            self._path = None
            self._address = (self._host, self._port)
            self._netloc = format_addr(self._address)

    @property
    def scheme(self):
        return self._scheme

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def path(self):
        return self._path

    @property
    def address(self):
        return self._address

    @property
    def netloc(self):
        return self._netloc

    def __eq__(self, other):
        return str(self) == str(other)

    def __str__(self):
        return '{}://{}'.format(self.scheme, self.netloc)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, str(self))



from __future__ import absolute_import

try: # py3
    import socketserver
    from http import client
    from http import server
except ImportError: # py2
    import httplib as client
    import BaseHTTPServer as server
    import SocketServer as socketserver

from .. import logs
from .. import utils
from .. import param
from .. import __version__

from . import Connection, Transport

SERVER_NAME = 'snekrpc'

log = logs.get(__name__)

# BaseRequestHandler is remarkably not an object in Py2
class HTTPHandler(server.BaseHTTPRequestHandler, object):
    protocol_version = 'HTTP/1.1'

    def do_POST(self):
        # handle command
        ifc = self.server._interface
        con = HTTPServerConnection(
            ifc, utils.url.format_addr(self.client_address), self)
        with con:
            while not con.should_close:
                ifc.handle(con)

    def start_response(self):
        # start response
        self.send_response(200)
        self.send_header('Connection', 'keep-alive')
        # TODO: set the content-type based on the codec used
        self.send_header('Content-Type', 'application/octet-stream')
        self.send_header('Transfer-Encoding', 'chunked')

        # user headers
        for k, v in self.server._headers.items():
            self.send_header(k, v)
        self.end_headers()

    def version_string(self):
        return self.server._version or '{}/{} {}'.format(SERVER_NAME,
            __version__, self.server._interface.version or '').strip()

    def log_request(self, code='-', size='-'):
        url = utils.url.format_addr(self.client_address)
        log.debug('%r %s <- %s', self.requestline, code, url)

class HTTPTransport(Transport):
    _name_ = 'http'
    Handler = HTTPHandler

    @param('headers', dict)
    def __init__(self, url, headers=None, version=None):
        super(HTTPTransport, self).__init__(url)
        url = self._url
        self._addr = (url.host, url.port)

        self.headers = headers
        self.version = version

    def connect(self, client):
        return HTTPClientConnection(client, utils.url.format_addr(self._addr))

    def serve(self, server):
        self._server = ThreadingHTTPServer(self._addr, self.Handler)
        self._server._headers = self.headers or {}
        self._server._version = self.version
        self._server._interface = server

        log.info('listening: %s', self.url)
        # XXX: expose poll_interval for config?
        self._server.serve_forever()

    def stop(self):
        # close the server's socket
        self._server.server_close()
        # stop the serve loop
        self._server.shutdown()

    def join(self, timeout=None):
        # ideally we could pass the timeout to the __is_shut_down Event here
        pass

class ThreadingHTTPServer(socketserver.ThreadingMixIn, server.HTTPServer):
    daemon_threads = True

    def handle_error(self, request, client_address):
        log.exception('request error (%s)', utils.url.format_addr(client_address))

class HTTPClientConnection(Connection):
    def __init__(self, client_ifc, url):
        super(HTTPClientConnection, self).__init__(client_ifc, url)

        self._res = None

        self._con = con = client.HTTPConnection(url)
        con.auto_open = False
        self.connect()

    def connect(self):
        con = self._con
        con.connect()
        log.debug('connected: %s', self.url)

        con.putrequest('POST', '')
        con.putheader('Connection', 'keep-alive')
        con.putheader('Content-Type', 'application/octet-stream')
        con.putheader('Transfer-Encoding', 'chunked')
        con.endheaders()

    def recv(self):
        if not self._res:
            self._res = self._con.getresponse()
        rfile = self._res.fp

        line = rfile.readline()
        chunk_len = int(line[:-2], 16)

        # +2 for the trailing newline (not included in chunk_len)
        return rfile.read(chunk_len + 2)[:-2]

    def send(self, data):
        con = self._con

        con.send('{:X}\r\n'.format(len(data)).encode('ascii'))
        con.send(data + b'\r\n')

    def close(self):
        log.debug('disconnected: %s', self.url)

class HTTPServerConnection(Connection):
    def __init__(self, server, url, handler):
        super(HTTPServerConnection, self).__init__(server, url)

        self.handler = handler
        self.response_started = False
        self.should_close = False

        log.debug('connected: %s', self.url)

    def recv(self):
        rfile = self.handler.rfile

        line = rfile.readline()
        if not line:
            self.should_close = True
            return None
        chunk_len = int(line[:-2], 16)

        # +2 for the trailing newline (not included in chunk_len)
        return rfile.read(chunk_len + 2)[:-2]

    def send(self, data):
        if not self.response_started:
            self.handler.start_response()
            self.response_started = True

        wfile = self.handler.wfile

        # send chunked
        wfile.write('{:X}\r\n'.format(len(data)).encode('ascii'))
        wfile.write(data + b'\r\n')

    def close(self):
        self.handler.wfile.flush()
        log.debug('disconnected: %s', self.url)



from . import logs
from . import utils
from . import errors
from . import registry
from .codec import get as get_codec
from .transport import get as get_transport
from .service import ServiceProxy, get as get_service

DEFAULT_CODEC = 'msgpack'

log = logs.get(__name__)

class Interface(object):
    def __init__(self, transport=None, codec=None, version=None):
        """Base class for client or server interfaces.

        *transport* can be a `transport.Transport` instance, or the name of a
        registered `transport.Transport` class.

        *codec* can be a `codec.Codec` instance, or the name of a
        registered `codec.Codec` class.

        *version* is an optional version string for tracking changes.
        """
        # initialize all metaclass registries
        registry.init()

        self.transport = get_transport(transport or utils.DEFAULT_URL)
        self.codec = codec
        self.version = version

    @property
    def url(self):
        return self.transport.url

    @property
    def codec(self):
        return self._codec

    @codec.setter
    def codec(self, codec):
        self._codec = codec and get_codec(codec)
        log.debug('codec: %s', self._codec and self._codec._name_)

class Client(Interface):
    def __init__(self, transport=None, codec=None, version=None,
            retry_count=None, retry_interval=None):
        super(Client, self).__init__(transport, codec, version)

        self._con = None

        self.retry_count = retry_count
        self.retry_interval = retry_interval

    ## connection ##

    def connect(self):
        if not self._con:
            try:
                self._con = self.transport.connect(self)
            except Exception as e:
                raise errors.TransportError(e)
        return self._con

    def close(self):
        if self._con:
            self._con.close()
        self._con = None

    ## remote services ##

    def __getitem__(self, name):
        """Convenience access to `service()`."""
        return self.service(name)
    
    def __getattr__(self, name):
        """Convenience access to `service()`.
        
        Note that this will not work if the service name is that same
        as a `Client` attribute. Use the `service()` method in those cases.
        """
        try:
            return self.service(name)
        except errors.RemoteError as e:
            if e.name != 'KeyError':
                raise
            raise AttributeError(name)

    def service(self, name, metadata=True):
        """Returns a `ServiceProxy` instance providing access to a remote
        service.

        If *metadata* is `True`, retrieves the metadata for the service commands
        from the remote interface. *metadata* can also be set directly as a list
        of command metadata.
        """
        return ServiceProxy(name, self, metadata)

    def service_names(self):
        return self.service('_meta').service_names()

class Server(Interface):
    def __init__(self, transport=None, codec=None, version=None,
            remote_tracebacks=False):
        super(Server, self).__init__(transport, codec or DEFAULT_CODEC, version)

        self._services = {}
        self.add_service('meta', {'server': self}, '_meta')

        self.remote_tracebacks = remote_tracebacks

    ## connection ##

    def serve(self):
        try:
            self.transport.serve(self)
        except Exception as e:
            raise errors.TransportError(e)

    def handle(self, con):
        con.get_protocol().handle()

    def stop(self):
        self.transport.stop()

    def join(self, timeout=None):
        self.transport.join(timeout)

    ## local services ##

    def add_service(self, service, service_args=None, alias=None):
        if service == 'meta': # special case
            service_args = {'server': self}

        svc = get_service(service, service_args, alias)
        self._services[svc._name_] = svc

        log.debug('service added: %s', svc._name_)

        return self

    def remove_service(self, name):
        del self._services[name]
        return self

    def service(self, name):
        return self._services[name]

    def services(self):
        return [self.service(n) for n in self.service_names()]

    def service_names(self):
        return [k for k in self._services if k and not k.startswith('_')]



import threading

from . import logs
from . import errors
from .utils.path import import_package

log = logs.get(__name__)

_metaclasses = {}
_init_lock = threading.Lock()
_initialized = threading.Event()

def init():
    """Import and register modules for all known metaclass registries.

    Safe to call multiple times from multiple threads.
    """
    with _init_lock:
        if _initialized.is_set():
            return

        for meta_name, Meta in _metaclasses.items():
            Meta.init(meta_name)

        _initialized.set()

def create_metaclass(meta_name):
    class RegistryMeta(type):
        registry = {}

        def __init__(cls, name, bases, dct):
            if bases[0] is not object:
                cls._name_ = reg_name = dct.get('_name_', name)
                if reg_name in RegistryMeta.registry:
                    raise errors.RegistryError(
                        'already registered: {}'.format(reg_name))
                else:
                    RegistryMeta.registry[reg_name] = cls
                    log.debug('registered %s: %s',
                        meta_name.split('.', 1)[1], reg_name)
            super(RegistryMeta, cls).__init__(name, bases, dct)

        @classmethod
        def get(cls, name):
            _initialized.wait()
            return cls.registry[name]

        @classmethod
        def names(cls):
            _initialized.wait()
            return tuple(cls.registry.keys())

        @classmethod
        def init(cls, name):
            exceptions = import_package(name)
            for modname, exc in exceptions.items():
                RegistryMeta.registry[modname] = exc

    _metaclasses[meta_name] = RegistryMeta
    return RegistryMeta



import time

from .. import logs
from .format import format_exc

RETRY_COUNT = 0
RETRY_INTERVAL = 1.0
RETRY_ERRORS = (Exception,)
RETRY_LOG_TEMPLATE = '{error} (retrying: {retries})'

log = logs.get(__name__)

class Retry(object):
    def __init__(self, count=None, interval=None, errors=None, logger=None,
            log_template=None):
        self.count = RETRY_COUNT if count is None else count
        self.interval = RETRY_INTERVAL if interval is None else interval
        self.errors = tuple(errors or RETRY_ERRORS)

        logger = logger or log
        self.log = (logger.exception
            if logger.isEnabledFor(logs.DEBUG) else log.warning)
        self.log_template = log_template or RETRY_LOG_TEMPLATE

    def call(self, func, *args, **kwargs):
        retries = 0
        count = self.count
        interval = self.interval
        errors = self.errors

        while True:
            try:
                return func(*args, **kwargs)
            except errors as e:
                if retries >= count >= 0:
                    raise

                time.sleep(interval)
                retries += 1

                self._log(retries, e)

    def call_gen(self, func, *args, **kwargs):
        retries = 0
        count = self.count
        interval = self.interval
        errors = self.errors

        while True:
            try:
                for x in func(*args, **kwargs):
                    yield x
                return
            except errors as e:
                if retries >= count >= 0:
                    raise

                time.sleep(interval)
                retries += 1

                self._log(retries, e)

    def _log(self, retries, exc):
        self.log(self.log_template.format(
            error=format_exc(exc),
            retries=retries,
            count=self.count,
            interval=self.interval,
            ))



import inspect
import itertools

from .. import logs
from .. import utils
from .. import errors
from .. import registry
from ..utils.encoding import to_unicode

log = logs.get(__name__)

def parse_alias(name):
    try:
        name, alias = name.split(':')
    except ValueError:
        alias = None
    return name, alias

def get_class(name):
    return ServiceMeta.get(name)

def get(name, service_args=None, alias=None):
    """Returns an instance of the Service matching *name*."""
    if isinstance(name, Service):
        obj = name
    elif inspect.isclass(name) and issubclass(name, Service):
        obj = name(**(service_args or {}))
    else:
        cls = get_class(name)
        obj = cls(**(service_args or {}))

    # rename the instance if any alias was set
    if alias:
        obj._name_ = alias

    return obj

def service_to_dict(svc):
    """Returns service metadata as a dict."""
    f2d = utils.function.func_to_dict

    d = {
        'name': svc._name_,
        'doc': svc.__doc__,
        }
    d['commands'] = cmds = []

    # find commands
    for name in dir(svc):
        if name.startswith('_'):
            continue
        attr = getattr(svc, name)
        meta = getattr(attr, '_meta', None)
        if meta is not None:
            cmds.append(f2d(attr))

    return d

ServiceMeta = registry.create_metaclass(__name__)

class Service(utils.compat.with_metaclass(ServiceMeta, object)):
    _name_ = None

    # this is here only to define an __init__ that takes no args
    # otherwise Python will use a `method-wrapper`, which can somehow produce
    # a `Signature` with different arguments than what it actually takes
    def __init__(self):
        pass

class ServiceProxy(object):
    """A proxy class providing access to a remote service.

    If *metadata* is `True`, the metadata service will be used to load metadata
    for the selected remote service. If `False`, no metadata will be loaded.
    It is also possible to use *metadata* to pass a metadata dict directly.
    """
    def __init__(self, name, client, metadata=True):
        self._svc_name = to_unicode(name)
        self._client = client
        self._commands = cmds = {}

        self._retry = utils.retry.Retry(
            client.retry_count, client.retry_interval,
            errors=[errors.TransportError], logger=log)

        # collect metadata. convert commands to functions
        wrap = lambda c: wrap_call(self, c['name'], c)
        if metadata is True:
            meta = ServiceProxy('_meta', client, metadata=False)
            svc = meta.service(self._svc_name)
            cmds.update({c['name']: wrap(c) for c in svc['commands']})
        elif metadata:
            cmds.update({c['name']: wrap(c) for c in metadata})

    def __getattr__(self, cmd_name):
        if self._commands:
            # return a metadata-based wrapper
            try:
                return self._commands[cmd_name]
            except KeyError:
                raise AttributeError(cmd_name)
        # return a raw wrapper
        return wrap_call(self, cmd_name)

    def __dir__(self):
        return list(self._commands.keys()) + super(ServiceProxy, self).__dir__()

def wrap_call(proxy, cmd_name, cmd_def=None):
    # Protocol.send_cmd() is wrapped by two functions here in order to simplify
    # connection cleanup for both regular and generator calls.
    # For that reason call() is a generator, and call_wrap() handles the
    # case where the command is not a generator
    def call(*args, **kwargs):
        # with proxy._client.connect() as con:
        con = proxy._client.connect()
        try:
            proto = con.get_protocol({proxy._svc_name: proxy._commands})

            res = proto.send_cmd(proxy._svc_name, to_unicode(cmd_name),
                *args, **to_unicode(kwargs, dict_keys_only=True))

            # notify call_wrap of type of response
            isgen = inspect.isgenerator(res)
            yield isgen

            if isgen:
                for r in res:
                    yield r
            else:
                yield res
        except errors.TransportError:
            proxy._client.close()
            raise

    def call_wrap(*args, **kwargs):
        gen = call(*args, **kwargs)
        isgen = next(gen) # get type of response
        return iter(StreamInitiator(gen)) if isgen else next(gen)

    if cmd_def and cmd_def.get('isgen'):
        retry_wrap = lambda *a, **k: proxy._retry.call_gen(call_wrap, *a, **k)
    else:
        retry_wrap = lambda *a, **k: proxy._retry.call(call_wrap, *a, **k)

    if not cmd_def:
        return retry_wrap
    return utils.function.dict_to_func(cmd_def, retry_wrap)

class StreamInitiator(object):
    """Captures the first value of a generator to ensure that it begins
    execution."""
    def __init__(self, gen):
        # cache first result
        try:
            gen = itertools.chain([next(gen)], gen)
        except StopIteration:
            # keep the generator to raise StopIteration later
            pass
        self._gen = gen

    def __iter__(self):
        for x in self._gen:
            yield x



from __future__ import absolute_import

import json
from . import Codec, encode, decode

class JsonCodec(Codec):
    _name_ = 'json'

    def __init__(self, encoding=None):
        self._encoding = encoding or 'utf8'

    def encode(self, msg):
        return json.dumps(msg, default=encode).encode(self._encoding)

    def decode(self, data):
        return json.loads(data, object_hook=decode)



from __future__ import absolute_import

import io
import ssl
import errno
import socket
import threading
import collections

from h2 import events as ev
from h2.config import H2Configuration
from h2.connection import H2Connection

from .. import logs
from .. import utils
from .. import param
from .. import __version__

from . import Connection, Transport

BACKLOG = socket.SOMAXCONN
CHUNK_SIZE = io.DEFAULT_BUFFER_SIZE
ACCEPT_TIMEOUT = 0.1
SERVER_NAME = 'snekrpc'

log = logs.get(__name__)

Request = collections.namedtuple('Request', 'headers data')

class HTTP2Connection(Connection):
    def __init__(self, interface, sock, url, chunk_size=None, client_side=True):
        super(HTTP2Connection, self).__init__(interface, url)
        self._sock = sock
        self._chunk_size = chunk_size

        cfg = H2Configuration(client_side=client_side)
        self._con = con = H2Connection(config=cfg)
        con.initiate_connection()
        sock.sendall(con.data_to_send())

        # the server needs to keep track of the request's stream_id
        self._stream_id = None
        self._events = []

        log.debug('connected: %s', self.url)

    def recv(self):
        """Reads from the socket and processes all HTTP/2 events.

        Returns only on EOF or DataRecieved events.
        """
        con = self._con
        sock = self._sock
        events = self._events

        while True:
            if not events:
                # recieve new events
                data = sock.recv(CHUNK_SIZE)
                if not data:
                    return
                events.extend(con.receive_data(data))

            # process current events
            while True:
                try:
                    event = events.pop(0)
                except IndexError:
                    break

                if isinstance(event, ev.DataReceived):
                    return event.data
                elif isinstance(event, ev.RequestReceived):
                    log.debug('%s', event)
                    self._stream_id = event.stream_id
                elif isinstance(event, ev.ResponseReceived):
                    log.debug('%s', event)
                elif isinstance(event, ev.StreamEnded):
                    log.debug('%s', event)
                    self._stream_id = None
                elif isinstance(event, ev.RemoteSettingsChanged):
                    pass
                elif isinstance(event, ev.SettingsAcknowledged):
                    pass
                else:
                    log.warning('unhandled event: %s', event)

    def close(self):
        sock = self._sock
        try:
            sock.shutdown(socket.SHUT_RDWR)
        except (OSError, socket.error) as e:
            # ignore if not connected
            if e.errno not in (errno.ENOTCONN,):# errno.EBADF):
                raise
        sock.close()
        log.debug('disconnected: %s', self.url)

class HTTP2ServerConnection(HTTP2Connection):
    def __init__(self, interface, sock, url, chunk_size=None):
        super(HTTP2ServerConnection, self).__init__(interface, sock, url,
            chunk_size, client_side=False)

        self._initiated = False

    def initiate(self):
        stream_id = self._stream_id
        if self._initiated == stream_id:
            return

        headers = [
            (':status', '200'),
            ('server', self.version()),
            ]

        con = self._con
        con.send_headers(stream_id, headers)
        self._sock.sendall(con.data_to_send())

        self._initiated = stream_id

    def send(self, data):
        self.initiate()

        stream_id = self._stream_id
        con = self._con
        con.send_data(stream_id, data)

        self._sock.sendall(con.data_to_send())

    def version(self):
        return '{}/{} {}'.format(
            SERVER_NAME, __version__, self._ifc.version)

    def close(self):
        con = self._con
        con.end_stream(self._stream_id)
        self._sock.sendall(con.data_to_send())
        super(HTTP2ServerConnection, self).close()

class HTTP2ClientConnection(HTTP2Connection):
    def __init__(self, interface, sock, url, chunk_size=None):
        super(HTTP2ClientConnection, self).__init__(interface, sock, url,
            chunk_size)

    def send(self, data):
        headers = [
            (':method', 'POST'),
            (':authority', 'localhost'),
            (':scheme', 'http'),
            (':path', '/')
            ]

        con = self._con
        stream_id = con.get_next_available_stream_id()
        con.send_headers(stream_id, headers)
        con.send_data(stream_id, data)

        self._sock.sendall(con.data_to_send())

class HTTP2Transport(Transport):
    _name_ = 'http2'

    @param('backlog', int, default=BACKLOG)
    @param('chunk_size', int, default=CHUNK_SIZE)
    @param('ssl_key', doc='server-side only')
    def __init__(self, url, timeout=None, backlog=None, chunk_size=None,
            ssl_cert=None, ssl_key=None):
        super(HTTP2Transport, self).__init__(url)
        url = self._url

        self._addr = (url.host, url.port)
        self._sock = None

        self.timeout = timeout
        self.backlog = backlog or BACKLOG
        self.chunk_size = chunk_size

        self._ssl_cert = ssl_cert
        self._ssl_key = ssl_key

        self._stop = threading.Event()
        self._stopped = threading.Event()

    def connect(self, client):
        s = socket.create_connection(self._addr, self.timeout)

        if self._ssl_cert:
            ctx = ssl.create_default_context()
            ctx.load_verify_locations(self._ssl_cert)
            hostname = socket.gethostbyaddr(self._addr[0])[0]
            s = ctx.wrap_socket(s, server_hostname=hostname)

        return HTTP2ClientConnection(client, s, self._url.netloc, self.chunk_size)

    def bind(self):
        self._sock = s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(ACCEPT_TIMEOUT)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(self._addr)
        s.listen(self.backlog)

    def serve(self, server):
        self.bind()
        log.info('listening: %s', self.url)

        s = self._sock

        ctx = None
        if self._ssl_cert:
            ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ctx.load_cert_chain(certfile=self._ssl_cert, keyfile=self._ssl_key)

        stop = self._stop
        stopped = self._stopped
        stop.clear()
        stopped.clear()

        try:
            while not stop.is_set():
                try:
                    sock, addr = s.accept()
                except socket.timeout:
                    continue
                sock.settimeout(self.timeout)

                try:
                    if ctx:
                        sock = ctx.wrap_socket(sock, server_side=True)
                except ssl.SSLError:
                    log.exception('ssl error')
                    continue

                utils.start_thread(self.handle, server, sock, addr)
        finally:
            stopped.set()

    def handle(self, server, sock, addr):
        addr = utils.url.format_addr(addr)
        con = HTTP2ServerConnection(server, sock, addr, self.chunk_size)
        with con:
            server.handle(con)

    def stop(self):
        self._stop.set()

    def join(self, timeout=None):
        self._stopped.wait(timeout)



from __future__ import print_function

import inspect
import importlib

from .. import utils
from .. import registry

FormatterMeta = registry.create_metaclass(__name__)

def get(name, **kwargs):
    """Returns an instance of the Formatter matching *name*."""
    if isinstance(name, Formatter):
        return name
    cls = load(name) if '.' in name else FormatterMeta.get(name)
    return cls(**kwargs)

def load(name):
    """Imports and returns the Formatter class matching *name*."""
    mod_name, cls_name = name.rsplit('.', 1)
    mod = importlib.import_module(mod_name)
    return getattr(mod, cls_name)

class Formatter(utils.compat.with_metaclass(FormatterMeta, object)):
    def process(self, res):
        if inspect.isgenerator(res):
            for r in res:
                self.print(r)
        else:
            self.print(res)

    def print(self, res):
        print(self.format(res))

    def format(self, res):
        return res



from . import Service, ServiceProxy
from .. import Client, param
from .. import logs

log = logs.get(__name__)

class RemoteService(Service, ServiceProxy):
    """Provides access to a remote service"""
    _name_ = 'remote'

    @param('transport')
    @param('codec')
    @param('version')
    @param('retry_count', int)
    @param('retry_interval', float)
    @param('kwargs', hide=True)
    def __init__(self, name, **kwargs):
        Service.__init__(self)
        ServiceProxy.__init__(self, name, Client(**kwargs))

        log.info('forwarding (%s): %s', name, self._client.url)



# encoding: utf-8

"""
Terminal utilities for the CLI.

These functions are meant to provide useful functionality for output modules.
(Set on the CLI or pack script with the --output-module flag.)
"""

from __future__ import unicode_literals

import math
import calendar
import datetime

try:
    import colorama
    from colorama import Fore, Back, Style
except ImportError:
    colorama = None
else:
    colorama.init()

##
## colors
##

_enable_color = bool(colorama)
def set_color_enabled(enabled=True):
    global _enable_color
    _enable_color = enabled

def color(s, fore=None, back=None, style=None):
    if _enable_color and colorama:
        fore = getattr(Fore, fore.upper()) if fore else ''
        back = getattr(Back, back.upper()) if back else ''
        style = getattr(Style, style.upper()) if style else ''
        reset = Style.RESET_ALL
    else:
        fore, back, style, reset = [''] * 4
    return '{fore}{back}{style}{s}{reset}'.format(
        fore=fore, back=back, style=style, s=s, reset=reset)

##
## text alignment
##

def align(s, width, alignment='left'):
    try:
        alignment = {'left': '<', 'center': '^', 'right': '>'}[alignment]
    except KeyError:
        raise ValueError('invalid alignment: {}'.format(alignment))
    return '{:{}{}}'.format(s, alignment, width)

def elide(s, width=None, align='right', ellipsis='…'):
    """Append ellipsis to *s* if longer than *width*.

    Defaults to terminal width.
    """
    width = width or get_terminal_size().columns
    if len(s) <= width: return s

    l = len(ellipsis)
    if align == 'right':
        return s[:width-l] + ellipsis
    elif align == 'left':
        return ellipsis + s[-width+l:]
    elif align == 'center':
        w_2 = width / 2
        l_2 = l / 2
        return s[:math.ceil(w_2)-math.ceil(l_2)] + ellipsis + s[-math.floor(w_2)+math.floor(l_2):]
    else:
        raise ValueError('invalid alignment choice: {}'.format(align))

##
## date/time
##

# from: http://stackoverflow.com/a/13287083
def utc_to_local(utc_dt):
    """Convert a UTC datetime object to local time."""
    # get integer timestamp to avoid precision lost
    timestamp = calendar.timegm(utc_dt.timetuple())
    local_dt = datetime.datetime.fromtimestamp(timestamp)
    assert utc_dt.resolution >= datetime.timedelta(microseconds=1)
    return local_dt.replace(microsecond=utc_dt.microsecond)

def to_formatted_localtime(dt):
    return utc_to_local(dt).strftime('%Y-%m-%d %H:%M:%S')

##
## terminal
##

# Backport support for Python 3.3's `shutil.get_terminal_size()`.
# Adapted from: https://pypi.python.org/pypi/backports.shutil_get_terminal_size
# The `shutil` version is used when available.

try:
    from shutil import get_terminal_size
except ImportError:
    import os
    import struct
    import collections

    terminal_size = collections.namedtuple("terminal_size", "columns lines")

    try:
        from ctypes import windll, create_string_buffer, WinError

        _handle_ids = {
            0: -10,
            1: -11,
            2: -12,
        }

        def _get_terminal_size(fd):
            handle = windll.kernel32.GetStdHandle(_handle_ids[fd])
            if handle == 0:
                raise OSError('handle cannot be retrieved')
            if handle == -1:
                raise WinError()
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(handle, csbi)
            if res:
                res = struct.unpack("hhhhHhhhhhh", csbi.raw)
                left, top, right, bottom = res[5:9]
                columns = right - left + 1
                lines = bottom - top + 1
                return terminal_size(columns, lines)
            else:
                raise WinError()

    except ImportError:
        import fcntl
        import termios

        def _get_terminal_size(fd):
            try:
                res = fcntl.ioctl(fd, termios.TIOCGWINSZ, b"\x00" * 4)
            except IOError as e:
                raise OSError(e)
            lines, columns = struct.unpack("hh", res)

            return terminal_size(columns, lines)

    def get_terminal_size(fallback=(80, 24)):
        """Get the size of the terminal window.
        For each of the two dimensions, the environment variable, COLUMNS
        and LINES respectively, is checked. If the variable is defined and
        the value is a positive integer, it is used.
        When COLUMNS or LINES is not defined, which is the common case,
        the terminal connected to sys.__stdout__ is queried
        by invoking os.get_terminal_size.
        If the terminal size cannot be successfully queried, either because
        the system doesn't support querying, or because we are not
        connected to a terminal, the value given in fallback parameter
        is used. Fallback defaults to (80, 24) which is the default
        size used by many terminal emulators.
        The value returned is a named tuple of type os.terminal_size.
        """
        # Try the environment first
        try:
            columns = int(os.environ["COLUMNS"])
        except (KeyError, ValueError):
            columns = 0

        try:
            lines = int(os.environ["LINES"])
        except (KeyError, ValueError):
            lines = 0

        # Only query if necessary
        if columns <= 0 or lines <= 0:
            try:
                size = _get_terminal_size(sys.__stdout__.fileno())
            except (NameError, OSError):
                size = terminal_size(*fallback)

            if columns <= 0:
                columns = size.columns
            if lines <= 0:
                lines = size.lines

        return terminal_size(columns, lines)



# encoding: utf-8

from __future__ import print_function

import os
import sys
import json
import stat
import argparse
import datetime

from . import logs
from . import utils
from . import errors
from . import registry
from . import interface
from . import codec, formatter, service, transport
from .utils.function import Param

DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
DATE_FORMAT = '%Y-%m-%d'
TIME_FORMAT = '%H:%M:%S'

COLLECTION_TYPES = frozenset(['list', 'dict'])

def main():
    """Convenient entry-point."""
    Parser().main()

class Parser(object):
    def __init__(self):
        # global args
        self.base_parser = argparse.ArgumentParser(add_help=False)
        self.add_global_args(self.base_parser)

    def main(self):
        """Processes command-line arguments and calls any selected command."""

        # temp parser to grab connection arguments
        parser = argparse.ArgumentParser(add_help=False, parents=[self.base_parser])
        parser.add_argument('-h', '--help', action='store_true',
            help='show this help message and exit')
        parser.add_argument('rest', nargs=argparse.REMAINDER,
            help=argparse.SUPPRESS)

        args, extra_args = parser.parse_known_args()

        logs.init(args.verbose)

        # import built-in modules
        registry.init()

        # import additional modules
        for imp in args.imports:
            utils.path.import_module(imp)

        if args.list:
            meta = {
                'codecs': codec.CodecMeta,
                'formatters': formatter.FormatterMeta,
                'services': service.ServiceMeta,
                'transports': transport.TransportMeta,
                }[args.list]

            # select an output formatter
            try:
                fmt = formatter.get(args.format)
            except Exception as e:
                if args.verbose: raise
                parser.error(e)

            fmt.process(meta.names())
            return

        # parser for transport and service args
        parser = argparse.ArgumentParser(parents=[self.base_parser])

        cls = transport.get(args.url)
        if isinstance(cls, Exception):
            self.add_transport_exception(parser, args.url.scheme, cls)
        else:
            self.add_transport_args(parser, cls)

        # add service arguments
        used_aliases = set()
        for name in sorted(args.services):
            name, alias = service.parse_alias(name)
            cls = service.get_class(name)

            alias = alias or cls._name_
            if alias in used_aliases:
                raise ValueError('duplicate service alias: {}'.format(alias))
            used_aliases.add(alias)

            self.add_service_args(parser, cls, alias)

        # help
        if args.help:
            parser.print_help()
            parser.exit()

        # collect transport and service args
        sub_args = parser.parse_args(extra_args)
        trn_args, svc_args = self.get_prefixed_args(sub_args)

        trn_name = args.url.scheme
        trn = transport.get(args.url, trn_args.get(trn_name))

        # start client or server
        if args.server_mode:
            if args.rest:
                parser.error('unrecognized arguments: {}'.format(' '.join(args.rest)))
            self.start_server(trn, args, svc_args)
        else:
            self.start_client(trn, parser, args)

    def start_client(self, trn, parser, args):
        client = interface.Client(trn,
            codec=args.codec,
            retry_count=args.retry_count,
            retry_interval=args.retry_interval,
            )

        # get service metadata
        try:
            meta = client.service('_meta')

            if args.version:
                # show server status and exit
                status = meta.status()
                self.print_status(status)
                parser.exit()

            svcs = meta.services()

        except errors.RemoteError as e:
            if not args.verbose and e.name == 'KeyError':
                parser.error('metadata service not available')
            raise

        except Exception as e:
            if args.verbose: raise
            err = '{}\nconnection required for help on remote services'
            parser.error(err.format(e))

        # add services
        svc_subs = parser.add_subparsers(title='remote services')

        for svc in sorted(svcs, key=lambda s: s['name']):
            svc_name = svc['name']
            svc_desc = self.get_help(svc['doc'])
            svc_parser = svc_subs.add_parser(svc_name,
                help=svc_desc, description=svc['doc'])
            svc_parser.set_defaults(svc_name=svc_name)

            # add service commands
            cmd_subs = svc_parser.add_subparsers(title='commands', dest='command')
            cmd_subs.required = True

            for cmd in svc['commands']:
                cmd_name = cmd['name']
                cmd_desc = self.get_help(cmd['doc'])
                cmd_parser = cmd_subs.add_parser(cmd_name,
                    help=cmd_desc, description=cmd['doc'])
                self.add_command_args(cmd_parser, cmd)
                cmd_parser.set_defaults(cmd_name=cmd_name, cmd_meta=cmd)

        # help
        if args.help:
            parser.print_help()
            parser.exit()
        elif not args.rest:
            parser.print_usage()
            parser.exit()

        # select an output formatter
        try:
            fmt = formatter.get(args.format)
        except Exception as e:
            if args.verbose: raise
            parser.error(e)

        # get the command arguments
        verbose = args.verbose
        args = parser.parse_args(args.rest)
        cmd_args, cmd_kwargs = self.get_command_args(args)

        # get the command function
        svc = client.service(args.svc_name, metadata=[args.cmd_meta])
        func = getattr(svc, args.cmd_name)

        # call the command
        try:
            res = func(*cmd_args, **cmd_kwargs)
        except errors.RemoteError as e:
            if verbose: raise
            parser.error(e)
        fmt.process(res)

    def start_server(self, trn, args, svc_args):
        # create server
        s = interface.Server(trn, codec=args.codec, version=args.server_version,
            remote_tracebacks=args.remote_tracebacks)

        # add services
        for name in args.services:
            name, alias = service.parse_alias(name)
            s_args = svc_args.get(name, {})
            s.add_service(name, s_args, alias)

        s.serve()

    ## get arguments ##

    def get_prefixed_args(self, args):
        pfx_args = {}

        for name, value in vars(args).items():
            try:
                prefix, cls_name, arg_name = name.split('_', 2)
            except ValueError:
                continue

            cls_args = pfx_args.setdefault(prefix, {})
            cls_args.setdefault(cls_name, {})[arg_name] = value

        return pfx_args.get('transport', {}), pfx_args.get('service', {})

    def get_command_args(self, args):
        """Processes the command-line arguments and returns the arguments to
        pass to the selected command."""
        cmd = args.cmd_meta

        cmd_args = []
        cmd_kwargs = {}

        for param in cmd['params']:
            name = param['name']
            kind = param['kind']

            if param.get('hide', False):
                continue

            arg = getattr(args, param['name'])

            if kind in {Param.POSITIONAL_ONLY, Param.POSITIONAL_OR_KEYWORD}:
                cmd_args.append(arg)
            elif kind == Param.VAR_POSITIONAL:
                cmd_args.extend(arg)
            elif kind == Param.VAR_KEYWORD:
                cmd_kwargs.update(arg or {})
            elif kind == Param.KEYWORD_ONLY:
                cmd_kwargs[name] = arg
            else:
                assert False

        return cmd_args, cmd_kwargs

    ## add arguments ##

    def add_service_args(self, parser, cls, alias):
        svc_parser = parser.add_argument_group(
            '{} service arguments'.format(alias))

        # add a prefix to every param
        cmd = utils.function.func_to_dict(cls.__init__, remove_self=True)
        for param in cmd['params']:
            param['name'] = '_'.join(['service', alias, param['name']])
            # force keyword-only params for clarity
            if param['kind'] != Param.VAR_KEYWORD:
                param['kind'] = Param.KEYWORD_ONLY

            param.setdefault('default', argparse.SUPPRESS)

        self.add_command_args(svc_parser, cmd, single_flags=False)

    def add_transport_args(self, parser, cls):
        ignored = set(['url', 'timeout'])
        trn_parser = parser.add_argument_group(
            '{} transport arguments'.format(cls._name_),
            'To see arguments for another transport, set the "--url" argument'
            )

        # add a prefix to every param
        cmd = utils.function.func_to_dict(cls.__init__, remove_self=True)
        for param in cmd['params']:
            if param['name'] in ignored:
                param['hide'] = True
            param['name'] = '_'.join(['transport', cls._name_, param['name']])
            # force keyword-only params for clarity
            if param['kind'] != Param.VAR_KEYWORD:
                param['kind'] = Param.KEYWORD_ONLY

        self.add_command_args(trn_parser, cmd, single_flags=False)

    def add_transport_exception(self, parser, name, exc):
        parser.add_argument_group(
            '{} transport arguments'.format(name),
            'failed to load transport: {}'.format(exc),
            )

    def add_command_args(self, parser, cmd, single_flags=True):
        is_option_arg = lambda p: (
            p['kind'] == Param.VAR_KEYWORD or 'default' in p)

        if single_flags:
            # keep track of used single char flags
            chars = set(u'h')
            # include single char arguments
            chars.update(p['name'] for p in cmd['params']
                if len(p['name']) == 1 and not is_option_arg(p))
        else:
            chars = None

        for param in cmd['params']:
            name = param['name']
            kind = param['kind']
            hint = param.get('hint')
            doc = param.get('doc')
            default = param.get('default', Param.empty)

            if param.get('hide', False):
                continue

            kwargs = {'metavar': name}

            if default is not Param.empty:
                kwargs['default'] = default

            if kind in {Param.KEYWORD_ONLY, Param.VAR_KEYWORD}: # **kwargs
                self.add_option_arg(parser, param, chars)

            elif 'default' in param: # args with defaults
                self.add_option_arg(parser, param, chars)

            else: # positional args
                if kind == Param.VAR_POSITIONAL: # *args
                    kwargs['nargs'] = '*'

                kwargs.update({
                    'type': self.get_converter(hint),
                    'help': self.get_argument_help(doc, hint, default),
                    })
                parser.add_argument(name, **kwargs)

    def add_option_arg(self, parser, param, chars=None):
        name = param['name']
        kind = param['kind']
        hint = param.get('hint')
        doc = param.get('doc')
        default = param.get('default')

        # use dashes instead of underscores for param names
        flag_name = name.replace('_', '-')

        # check for possible short flags
        flags = []
        added = False
        c = flag_name[0]
        C = c.upper()

        # check if the lower or uppercase char is unique
        if not chars:
            # don't add a short flag
            pass
        elif c not in chars:
            flags.append('-' + c)
            chars.add(c)
            added = True
        elif C not in chars:
            flags.append('-' + C)
            chars.add(C)
            added = True

        # add a long flag if no short flag was added
        # add a long flag if the name is more than 1 character
        if not added or len(flag_name) > 1:
            flags.append('--' + flag_name)

        if hint == 'bool':
            # handle bool special case
            group = parser.add_mutually_exclusive_group()
            if not default:
                # in case default is None
                default = False

            help = self.get_argument_help(doc)

            # add a flag for the True value
            group.add_argument(*flags, action='store_true', default=default,
                dest=name, help=help + ' (default)' if default is True else '')

            # add a flag for the False value
            group.add_argument('--no-' + flag_name, action='store_false',
                dest=name, help=help + ' (default)' if default is False else '')

        elif kind == Param.VAR_KEYWORD:
            parser.add_argument(*flags,
                action='append', dest=name, type=self.get_converter('keyword'),
                metavar='name=value', default=default,
                help=self.get_argument_help(doc, None, default))

        else:
            parser.add_argument(*flags,
                dest=name, type=self.get_converter(hint),
                metavar=self.get_argument_hint(hint), default=default,
                help=self.get_argument_help(doc, None, default))

    def add_global_args(self, parser):
        """Adds an argument group to *parser* for global arguments."""

        egroup = parser.add_mutually_exclusive_group()
        egroup.add_argument('-C', '--client', action='store_false',
            dest='server_mode', help='start in client mode (default)',
            default=False)
        egroup.add_argument('-S', '--server', action='store_true',
            dest='server_mode', help='start in server mode')

        parser.add_argument('-l', '--list',
            choices=['codecs', 'formatters', 'services', 'transports'],
            help='list the modules available for the selected category')
        parser.add_argument('-v', '--verbose', action='count',
            default=0, help='enable verbose output (-vv for more)')
        parser.add_argument('-V', '--version', action='store_true',
            help='show server version')

        group = parser.add_argument_group('configuration arguments')

        group.add_argument('-i', '--import', action='append', dest='imports',
            metavar='IMPORT', default=[],
            help='import an additional (codec/formatter/service/transport) module')
        group.add_argument('-c', '--codec',
            help='the codec format to use (default: {} on server)'.format(
                interface.DEFAULT_CODEC))
        group.add_argument('-s', '--service', action='append', dest='services',
            metavar='SERVICE', default=[],
            help='add a service module (can be set multiple times) '
                'formats: "<built-in service>[:alias]" or '
                '"<module.ServiceClass>[:alias]"')

        group = parser.add_argument_group('client arguments')

        group.add_argument('-u', '--url', type=utils.url.Url,
            default=utils.DEFAULT_URL,
            help='URL to RPC server (default: {})'.format(utils.DEFAULT_URL))
        group.add_argument('-t', '--timeout', type=float,
            help='number of seconds to wait for a response (default: no timeout)')
        group.add_argument('-r', '--retry-count', type=int,
            help='number of retry attempts to make (-1 for unlimited, default: no retries)')
        group.add_argument('--retry-interval', type=float,
            help='number of seconds between retry attempts (default: 1.0)')

        group = parser.add_argument_group('server arguments')

        group.add_argument('--server-version',
            help='version string for the server')
        group.add_argument('--remote-tracebacks', action='store_true',
            help='send tracebacks with errors')

        group = parser.add_argument_group('output arguments')

        format_default = {
            'piped': 'json',
            'redirected': 'raw',
            'terminal': 'pretty',
            }[io_stat_mode()]
        group.add_argument('-f', '--format',
            default=format_default,
            help="select a formatter by name or provide a custom Formatter "
                "subclass (default: 'pretty' on terminals, "
                "'json' when piped, and 'raw' when redirected)")

    ## parser help ##

    def get_help(self, doc):
        doc = doc or '\n'
        return doc.splitlines()[0]

    def get_argument_help(self, doc=None, hint=None, default=None):
        if hint == 'stream':
            hint = "path or '-' for stdin"
        help = '<{}>'.format(hint) if hint else ''
        if doc:
            help = '{} {}'.format(doc, help)
        if default not in [Param.empty, argparse.SUPPRESS, None]:
            help += ' (default: {})'.format(default)
        return help

    def get_argument_hint(self, hint):
        if not hint:
            return '<str>'
        elif hint in COLLECTION_TYPES:
            return '<path or JSON>'
        else:
            return '<{}>'.format(hint)

    ## parser utils ##

    def get_converter(self, hint):
        """Returns a type converter keyed to a specific typehint."""
        if hint == 'int':
            conv = lambda v: int(v)
        elif hint == 'float':
            conv = lambda v: float(v)
        elif hint == 'bytes':
            conv = lambda v: utils.encoding.to_bytes(v)
        elif hint == 'stream':
            conv = lambda v: utils.path.iter_file(argparse.FileType('rb')(v))
        elif hint == 'keyword':
            conv = lambda v: v.split('=', 1)
        elif hint == 'datetime':
            def conv(v):
                try:
                    return datetime.datetime.strptime(v, DATETIME_FORMAT)
                except ValueError:
                    try:
                        return datetime.datetime.strptime(v, DATE_FORMAT)
                    except Exception:
                        return datetime.datetime.combine(
                            datetime.date.today(),
                            datetime.datetime.strptime(v, TIME_FORMAT).time(),
                            )
        elif hint == 'stream':
            conv = lambda v: (x for x in v)
        elif hint in COLLECTION_TYPES:
            def conv(v):
                try:
                    with open(v) as f:
                        return json.load(f)
                except Exception:
                    return json.loads(v)
        else:
            conv = lambda v: utils.encoding.to_unicode(v)

        # the converter name is used in error messages
        conv.__name__ = utils.function.get_func_name(hint or 'str')

        return conv

    def print_status(self, status):
        """Prints the server version and status data."""
        # TODO: abide by --format flag
        width = max(len(k) for k in status)
        for k, v in sorted(status.items()):
            print('{:>{}}: {}'.format(k, width, v))

##
## cli utils
##

def io_stat_mode():
    mode = os.fstat(sys.stdout.fileno()).st_mode
    if stat.S_ISFIFO(mode):
        return 'piped'
    elif stat.S_ISREG(mode):
        return 'redirected'
    else:
        return 'terminal'

# fix for open issue: https://bugs.python.org/issue14156
class FileType(argparse.FileType):
    def __call__(self, string):
        fp = super(FileType, self).__call__(string)
        if string == '-' and 'r' in self._mode:
            fp = getattr(fp, 'buffer', fp)
        return fp



from __future__ import absolute_import

import socket

from .. import logs
from .. import utils
from .. import param

from . import tcp

ACCEPT_TIMEOUT = 0.1

log = logs.get(__name__)

class UnixConnection(tcp.TcpConnection):
    log = log

class UnixTransport(tcp.TcpTransport):
    _name_ = 'unix'
    log = log
    Connection = UnixConnection

    @param('backlog', int, default=tcp.BACKLOG)
    @param('chunk_size', int, default=tcp.CHUNK_SIZE)
    def __init__(self, url, timeout=None, backlog=None, chunk_size=None):
        super(UnixTransport, self).__init__(url, timeout, backlog, chunk_size)
        self._path = self._url.path

    def connect(self, client):
        self._sock = s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        s.settimeout(self.timeout)
        s.connect(self._path)
        return self.Connection(client, s, self._path, self.chunk_size)

    def bind(self):
        # remove existing socket
        utils.path.discard_file(self._path)

        self._sock = s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        s.settimeout(ACCEPT_TIMEOUT)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(self._path)
        s.listen(self.backlog)

    def serve(self, server):
        try:
            super(UnixTransport, self).serve(server)
        finally:
            utils.path.discard_file(self._path)

    def handle(self, server, sock, addr):
        with self.Connection(server, sock, self._path, self.chunk_size) as con:
            server.handle(con)



def test_null(service):
    assert service.null() is None

def test_echo(service):
    value = 'echooo'
    assert service.echo(value) == value

def test_params(service):
    result = service.params(1)
    assert result == [1, None, True, False, [], {}]

    result = service.params(2, 'a', False, True, 1, 2, 3, z=1)
    assert result == [2, 'a', False, True, [1, 2, 3], {'z': 1}]

def test_upstream(service):
    rtup = list(range(10))

    # send generator
    result = service.upstream((x for x in range(10)))
    assert result == rtup

def test_downstream(service):
    result = list(service.downstream(limit=10))
    assert result == list(range(10))

def test_dualstream(service):
    rtup = list(range(10))
    result = list(service.dualstream(rtup))
    assert result == rtup



import io
import os
import glob
import shutil
import contextlib

from .. import logs
from .. import utils
from .. import Service, command, param

log = logs.get(__name__)

CHUNK_SIZE = io.DEFAULT_BUFFER_SIZE

class FileService(Service):
    """Provides file access commands"""
    _name_ = 'file'

    @param('root_path', doc='root path for all file operations')
    @param('safe_root', bool,
        doc='ensures that no file operation escapes the root path')
    @param('chunk_size', int, default=CHUNK_SIZE,
        doc='size of data to buffer when reading')
    def __init__(self, root_path=None, safe_root=True, chunk_size=None):
        self.root_path = root_path or os.getcwd()
        self.safe_root = safe_root
        self.chunk_size = chunk_size or CHUNK_SIZE

    @property
    def root_path(self):
        return self._root_path

    @root_path.setter
    def root_path(self, root_path):
        # expand/normalize
        path = os.path.expandvars(os.path.expanduser(root_path))
        path = os.path.abspath(os.path.normpath(path))
        # we join '' to ensure a trailing slash
        # this protects against some path traversals at the
        # startswith(root_path) check
        self._root_path = os.path.join(path, '')

    @command(with_metadata='bool')
    def paths(self, pattern=None, with_metadata=False):
        """Iterates over path names matching the recursive glob *pattern*.

        Extended glob syntax is supported, so e.g. "**/*" can be used for
        recursion.

        If *with_metadata* is `True`, this command returns tuples of path
        metadata: `(path, file_size, modified_time, is_file)`.

        Only the path will be set in the tuple if the metadata cannot be read.
        """
        pattern = self.check_path(pattern or '*')
        if os.path.isdir(pattern):
            # match all files if pattern is a dir
            pattern = os.path.join(pattern, '*')

        for name in glob.iglob(pattern):
            if isinstance(name, bytes): # python 2
                name = name.decode('utf8')

            path = (os.path.relpath(name, self.root_path)
                if self.safe_root else name)
            if path == '.':
                continue

            if not with_metadata:
                yield path
                continue

            entry = {'path': path}

            try:
                stat = os.stat(name)
                isfile = os.path.isfile(name)

                entry.update(
                    size=stat.st_size,
                    mtime=stat.st_mtime,
                    isfile=isfile,
                    )
            except OSError as e:
                log.warning('could not read file: %s', e)
                entry.update(entry.fromkeys(['size', 'mtime', 'isfile']))

            yield utils.encoding.to_unicode(entry, dict_keys_only=True)

    @command()
    def touch(self, path):
        """Creates or updates the timestamps of a file at *path*."""
        path = self.check_path(path)
        with open(path, 'a'):
            os.utime(path, None)

    @command(recurse='bool')
    def create_dir(self, path, mode=None, recurse=False):
        """Create a directory at *path*.

        *mode* sets the directory permissions in octal (default: 755).

        If *recurse* is `True` existing directories will be ignored, and
        non-existing directories will be created as needed.

        Raises any exceptions that occur.
        """
        path = self.check_path(path)
        mode = 0o755 if mode is None else int(mode, 8)
        if recurse:
            utils.path.ensure_dirs(path, mode)
        else:
            os.mkdir(path, mode)

    @command(recurse='bool')
    def delete(self, path, recurse=False):
        """Deletes the file at *path*.

        If *recurse* is `True`, all children will also be deleted, including
        non-empty directories.

        Raises any exceptions that occur.
        """
        path = self.check_path(path)
        if os.path.isdir(path):
            if recurse:
                shutil.rmtree(path)
            else:
                os.rmdir(path)
        else:
            os.remove(path)

    @param('src', 'stream')
    @param('dst_path', doc='the remote path to upload to')
    @command()
    def upload(self, src, dst_path):
        """Uploads a file to the remote interface"""
        # first check access to the file
        path = self.check_path(dst_path)

        with open(path, 'wb') as fp:
            for chunk in src:
                fp.write(chunk)

            # ensure data is written to disk
            fp.flush()
            os.fsync(fp.fileno())

            log.info('%s bytes uploaded: %s', fp.tell(), path)

        # return the path that was used
        return self.sanitize_path(path)

    @param('src_path', doc='the remote path to download from')
    @command()
    def download(self, src_path):
        """Downloads a file from the remote interface"""
        # check access to the file
        path = self.check_path(src_path)

        with open(path, 'rb') as fp:
            for chunk in utils.path.iter_file(fp, self.chunk_size):
                yield chunk

            log.info('%s bytes downloaded: %s', fp.tell(), path)

    @command()
    def size(self, path):
        """Returns the contents of the file at *path*."""
        with self.sanitize_errors():
            path = self.check_path(path)
            return os.stat(path).st_size

    ## utils ##

    @contextlib.contextmanager
    def sanitize_errors(self):
        try:
            yield
        except OSError as e:
            e.filename = self.sanitize_path(e.filename)
            raise e

    def sanitize_path(self, path, root_path=None):
        """Returns only the public portion of the path.

        If `FileService.safe_root` is `False`, then just return the *path*
        """
        if not self.safe_root:
            return path
        return os.path.relpath(path, root_path or self.root_path)

    def check_path(self, path, root_path=None):
        """Ensures that *path* is a child of *root_path*.

        Returns *path* joined to *root_path*.

        If `FileService.safe_root` is `False`, the home dir (`~`) and
        environment variables will be expanded. Otherwise, an
        `IOError("Permission denied")` is thrown if the path does not resolve
        to a child of *root_path*.

        If *root_path* is not set, `FileService.root_path` will be used (CWD
        by default).
        """
        path = path or '.'
        root_path = root_path or self.root_path

        if not self.safe_root:
            path = os.path.expandvars(os.path.expanduser(path))
            return os.path.join(root_path, path)

        if not root_path:
            # user would have to manually set to null to get here
            raise ValueError('safe_root is enabled, but no root_path is set')

        # parent dir traversal is prohibited in safe_root mode
        if '..' in path:
            raise IOError("Permission denied: '{}'".format(path))

        # absolute path
        abs_path = os.path.join(root_path, path)

        # path has to start with root
        if not abs_path.startswith(root_path):
            raise IOError("Permission denied: '{}'".format(path))

        return abs_path



import io
import sys

PY3 = sys.version_info.major >= 3
# use str for unicode data and bytes for binary data
str = str if PY3 else unicode

def is_fileobject(o):
    return isinstance(o, io.IOBase if PY3 else file)

##
## Python 2/3 compatibility
## Pulled from the "future" library.
##

def with_metaclass(meta, *bases):
    """
    Function from jinja2/_compat.py. License: BSD.

    Use it like this::

        class BaseForm(object):
            pass

        class FormType(type):
            pass

        class Form(with_metaclass(FormType, BaseForm)):
            pass

    This requires a bit of explanation: the basic idea is to make a
    dummy metaclass for one level of class instantiation that replaces
    itself with the actual metaclass.  Because of internal type checks
    we also need to make sure that we downgrade the custom metaclass
    for one level to something closer to type (that's why __call__ and
    __init__ comes back from type etc.).

    This has the advantage over six.with_metaclass of not introducing
    dummy classes into the final MRO.
    """
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__
        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return metaclass('temporary_class', None, {})

def viewitems(obj):
    return getattr(obj, "viewitems", obj.items)()

def viewkeys(obj):
    return getattr(obj, "viewkeys", obj.keys)()

def viewvalues(obj):
    return getattr(obj, "viewvalues", obj.values)()



import pytest

from snekrpc.utils.url import Url, DEFAULT_HOST, DEFAULT_PORT

TEST_HOST = 'testhost'
TEST_PORT = 32123

def test_copy():
    url = 'tcp://{}:{}'.format(TEST_HOST, TEST_PORT)

    u1 = Url(url)
    u2 = Url(u1)

    assert u1 == u2 == url

def test_tcp():
    u = Url('tcp://{}:{}'.format(TEST_HOST, TEST_PORT))

    assert u.scheme == 'tcp'
    assert u.host == TEST_HOST
    assert u.port == TEST_PORT
    assert u.path is None
    assert u.address == (TEST_HOST, TEST_PORT)
    assert u.netloc == '{}:{}'.format(TEST_HOST, TEST_PORT)

def test_http():
    u = Url('http://{}:{}'.format(TEST_HOST, TEST_PORT))

    assert u.scheme == 'http'
    assert u.host == TEST_HOST
    assert u.port == TEST_PORT
    assert u.path is None
    assert u.address == (TEST_HOST, TEST_PORT)
    assert u.netloc == '{}:{}'.format(TEST_HOST, TEST_PORT)

def test_unix_relative():
    path = 'local/path/socket'
    u = Url('unix://{}'.format(path))

    assert u.scheme == 'unix'
    assert u.host is None
    assert u.port is None
    assert u.path == path
    assert u.address == path
    assert u.netloc == path

def test_unix_absolute():
    path = '/root/path/socket'
    u = Url('unix://{}'.format(path))

    assert u.scheme == 'unix'
    assert u.host is None
    assert u.port is None
    assert u.path == path
    assert u.address == path
    assert u.netloc == path

def test_unix_host_only():
    path = 'local'
    u = Url('unix://{}'.format(path))

    assert u.scheme == 'unix'
    assert u.host is None
    assert u.port is None
    assert u.path == path
    assert u.address == path
    assert u.netloc == path

def test_host_only():
    u = Url(TEST_HOST)

    assert u.scheme == 'tcp'
    assert u.host == TEST_HOST
    assert u.port == DEFAULT_PORT
    assert u.path is None
    assert u.address == (TEST_HOST, DEFAULT_PORT)
    assert u.netloc == '{}:{}'.format(TEST_HOST, DEFAULT_PORT)

def test_netloc_only():
    u = Url('{}:{}'.format(TEST_HOST, TEST_PORT))

    assert u.scheme == 'tcp'
    assert u.host == TEST_HOST
    assert u.port == TEST_PORT
    assert u.path is None
    assert u.address == (TEST_HOST, TEST_PORT)
    assert u.netloc == '{}:{}'.format(TEST_HOST, TEST_PORT)

def test_port_only():
    u = Url(':32123')

    assert u.scheme == 'tcp'
    assert u.host == DEFAULT_HOST
    assert u.port == TEST_PORT
    assert u.path is None
    assert u.address == (DEFAULT_HOST, TEST_PORT)
    assert u.netloc == '{}:{}'.format(DEFAULT_HOST, TEST_PORT)

def test_asterisk_only():
    u = Url('*')

    assert u.scheme == 'tcp'
    assert u.host == '0.0.0.0'
    assert u.port == DEFAULT_PORT
    assert u.path is None
    assert u.address == ('0.0.0.0', DEFAULT_PORT)
    assert u.netloc == '{}:{}'.format('0.0.0.0', DEFAULT_PORT)

def test_asterisk():
    u = Url('tcp://*:{}'.format(TEST_PORT))

    assert u.scheme == 'tcp'
    assert u.host == '0.0.0.0'
    assert u.port == TEST_PORT
    assert u.path is None
    assert u.address == ('0.0.0.0', TEST_PORT)
    assert u.netloc == '{}:{}'.format('0.0.0.0', TEST_PORT)

def test_trailing_slash():
    u = Url('{}:{}/'.format(TEST_HOST, TEST_PORT))

    assert u.scheme == 'tcp'
    assert u.host == TEST_HOST
    assert u.port == TEST_PORT
    assert u.path is None
    assert u.address == (TEST_HOST, TEST_PORT)
    assert u.netloc == '{}:{}'.format(TEST_HOST, TEST_PORT)

def test_to_str():
    url = 'tcp://{}:{}'.format(TEST_HOST, TEST_PORT)
    u = Url(url)
    assert url == str(u)

    url = 'unix://local/path/socket'
    u = Url(url)
    assert url == str(u)

    url = 'unix:///root/path/socket'
    u = Url(url)
    assert url == str(u)

def test_invalid():
    with pytest.raises(ValueError):
        Url('xxx:/asdf')
    with pytest.raises(ValueError):
        Url('tcp://asdf/asdf')



class Op:
    handshake    = 0 # <raw> initial handshake (codec)
    command      = 1 # call a command (svc_name, cmd_name, args, kwargs)
    data         = 2 # return data (data)
    error        = 3 # return error (name, msg, tb)
    stream_start = 4 # start of stream ([name])
    stream_end   = 5 # end of stream

    @classmethod
    def to_str(cls, op):
        for k, v in vars(cls).items():
            if v == op:
                return k
        raise ValueError('invalid op code: {}'.format(op))



import time
import itertools

from .. import Service, command

class HealthService(Service):
    """Provides basic connectivity commands"""
    _name_ = 'health'

    @command()
    def ping(self, count=1, interval=1.0):
        it = range(count-1) if count > 0 else itertools.count()
        for _ in it:
            yield
            time.sleep(interval)



import itertools

import snekrpc

class TestService(snekrpc.Service):
    _name_ = 'test'

    @snekrpc.command()
    def null(self):
        return None

    @snekrpc.command()
    def echo(self, s):
        return s

    @snekrpc.command()
    def params(self, a, b=None, c=True, d=False, *e, **f):
        return a, b, c, d, e, f

    @snekrpc.command(stream='stream')
    def upstream(self, stream):
        return list(stream)

    @snekrpc.command(limit='int')
    def downstream(self, limit=None):
        it = iter(range(limit)) if limit else itertools.count(0)
        while True:
            try:
                yield next(it)
            except StopIteration:
                return

    @snekrpc.command(stream='stream')
    def dualstream(self, stream):
        for x in stream:
            yield x



class SnekRPCError(Exception):
    """Base class for all errors."""

## transport

class TransportError(SnekRPCError):
    """Raised for any error in the transport."""

## io

class SendInterrupted(TransportError):
    """Raised when data sent to the remote end is less than expected."""

class ReceiveInterrupted(TransportError):
    """Raised when data received from the remote end is less than expected."""

## client

class ClientError(SnekRPCError):
    """Base class for client exceptions."""

# TODO: use these
class InvalidService(ClientError):
    """Raised for any attempt to access a service that does not exist."""

class InvalidCommand(ClientError):
    """Raised for any attempt to access a command that does not exist."""

class RemoteError(ClientError):
    """Raised for any exceptions that occur on the RPC server."""
    def __init__(self, name, msg, traceback):
        self.name = name
        self.msg = msg
        self.traceback = traceback

    @property
    def message(self):
        return '{}: {}'.format(self.name, self.msg)

    def __str__(self):
        return self.traceback or self.message

## server

class ServerError(SnekRPCError):
    """Base class for server exceptions."""

class ParameterError(ServerError):
    """Raised for invalid paramater configurations."""

## protocol

class ProtocolOpError(SnekRPCError):
    """Raised for protocol errors."""
    def __init__(self, opcode):
        super(ProtocolOpError, self).__init__('invalid opcode: {}'.format(opcode))
        self.opcode = opcode

## serialization

class EncodeError(SnekRPCError):
    """Adds context for errors raised when packing."""

class DecodeError(SnekRPCError):
    """Adds context for errors raised when unpacking."""

## registry

class RegistryError(SnekRPCError):
    pass



try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

from .compat import str, viewitems

def to_bytes(s, encoding='utf8'):
    if isinstance(s, bytes):
        pass
    elif isinstance(s, str):
        s = s.encode(encoding)
    elif isinstance(s, dict):
        s = {to_bytes(k, encoding): to_bytes(v, encoding)
            for k, v in viewitems(s)}
    elif isinstance(s, Iterable):
        s = [to_bytes(x, encoding) for x in s]
    return s

def to_unicode(s, encoding='utf8', dict_keys_only=False):
    """Attempts to decode all encountered bytes to unicode.

    *s* can be any standard data structure.

    If *dict_keys_only* is `True`, only the keys of a dict will be decoded.
    Values will remain untouched.
    """
    if isinstance(s, bytes):
        s = s.decode(encoding)
    elif isinstance(s, str):
        pass
    elif isinstance(s, dict):
        s = {to_unicode(k, encoding):
            v if dict_keys_only else to_unicode(v, encoding)
            for k, v in viewitems(s)}
    elif isinstance(s, Iterable):
        s = tuple(to_unicode(x, encoding) for x in s)
    return s



import traceback

def format_cmd(svc_name, cmd_name, args, kwargs):
    argstr = []
    if args:
        argstr.append(', '.join([repr(v) for v in args]))
    if kwargs:
        argstr.append(', '.join('{}={}'.format(k, repr(v))
            for k, v in kwargs.items()))
    return '{}.{}({})'.format(svc_name, cmd_name, elide(', '.join(argstr)))

def format_exc(exc):
    return traceback.format_exception_only(exc.__class__, exc)[0].strip()

# XXX: term.elide ?
def elide(s, width=100):
    return s if len(s) <= width else s[:width-3] + '...'



import os
from setuptools import setup, find_packages

def get_version():
    path = os.path.join(os.path.dirname(__file__), 'snekrpc/__init__.py')
    with open(path) as f:
        for line in f:
            if line.startswith('__version__'):
                return line.split('=')[1].strip().strip("'")
        return '0.0.0'

def get_readme():
    with open('README.md') as f:
        return f.read()

setup(
    name='snekrpc',
    version=get_version(),
    url='https://github.com/dhagrow/snekrpc',
    author='Miguel Turner',
    author_email='cymrow@gmail.com',
    description='A Python RPC library for prototyping.',
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['test']),
    entry_points={
        'console_scripts': [
            'snekrpc=snekrpc.__main__:main',
            'snekrpc-pack=snekrpc.pack:main',
            ],
    },
    install_requires=[
        'msgpack',
        'temporenc',
        ],
    )



from __future__ import absolute_import

import msgpack
from . import Codec, encode, decode

class MsgpackCodec(Codec):
    _name_ = 'msgpack'

    def encode(self, msg):
        return msgpack.packb(msg, use_bin_type=True, default=encode)

    def decode(self, data):
        return msgpack.unpackb(data, use_list=True, raw=False, object_hook=decode)



import inspect
import datetime

import temporenc

from .. import errors
from .. import utils
from .. import registry

CodecMeta = registry.create_metaclass(__name__)

def get(name, codec_args=None):
    """Returns an instance of the Codec matching *name*."""
    if isinstance(name, Codec):
        return name
    cls = CodecMeta.get(name)
    return cls(**(codec_args or {}))

class Codec(utils.compat.with_metaclass(CodecMeta, object)):
    _name_ = None

    def encode(self, msg):
        raise NotImplementedError('abstract')

    def decode(self, data):
        raise NotImplementedError('abstract')

    def _encode(self, msg):
        """Wrapper that provides encoding error context. Used internally."""
        try:
            return self.encode(msg)
        except Exception as e:
            raise errors.EncodeError('{}: msg={}'.format(e,
                utils.format.elide(repr(msg))))

    def _decode(self, data):
        """Wrapper that provides decoding error context. Used internally."""
        try:
            return self.decode(data)
        except Exception as e:
            raise errors.DecodeError('{}: data={!r}'.format(e,
                utils.format.elide(repr(data))))

## combined codecs

def encode(obj):
    if isinstance(obj, datetime.datetime):
        return encode_datetime(obj)
    elif inspect.isgenerator(obj):
        return encode_generator(obj)
    return obj

def decode(obj):
    if u'__datetime__' in obj:
        return decode_datetime(obj)
    elif u'__generator__' in obj:
        return decode_generator(obj)
    return obj

## datetime support

def encode_datetime(obj):
    return {u'__datetime__': temporenc.packb(obj)}

def decode_datetime(obj):
    return temporenc.unpackb(obj[u'__datetime__']).datetime()

## generator support
## Note that this is only intended to mark that *obj* is a generator. It
## does not encode/decode the contents.
## This is important for the argument handling in `protocol.py:recv_cmd()`

def encode_generator(obj):
    return {u'__generator__': None}

def decode_generator(obj):
    # any generator is fine
    return (x for x in ())



import sys
from logging import *

get = getLogger
log = get(__name__)

def init(debug_level=0, log_exceptions=True):
    """Initializes simple logging defaults."""
    root_log = get()

    # init only once
    if root_log.handlers: return

    fmt = '%(levelname).1s %(asctime)s . %(message)s'
    formatter = Formatter(fmt)

    handler = StreamHandler()
    handler.setFormatter(formatter)

    root_log.addHandler(handler)
    root_log.setLevel(DEBUG if debug_level > 0 else INFO)

    log = get('snekrpc.transport')
    log.setLevel(DEBUG if debug_level > 1 else INFO)

    log = get('snekrpc.utils')
    log.setLevel(DEBUG if debug_level > 1 else INFO)

    if log_exceptions:
        sys.excepthook = handle_exception

    # 3rd-party logs
    get('hpack').setLevel(INFO)

def handle_exception(etype, evalue, etb):
    if issubclass(etype, KeyboardInterrupt):
        sys.__excepthook__(etype, evalue, etb)
        return
    log.error('unhandled exception', exc_info=(etype, evalue, etb))



from __future__ import print_function

import pprint

from . import Formatter
from ..utils import compat

str = compat.str

class PrettyFormatter(Formatter):
    _name_ = 'pretty'

    def print(self, res):
        s = self.format(res)
        if s is not None:
            print(s)

    def format(self, res):
        if res is None:
            return
        elif isinstance(res, str):
            return res
        elif isinstance(res, (list, tuple)):
            return '\n'.join(str(r) for r in res)
        else:
            return pprint.pformat(res)



import io
import os
import errno
import pkgutil
import importlib

from .. import logs

log = logs.get(__name__)

BASE_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))

def base_path(*names):
    return os.path.join(BASE_PATH, *names)

def ensure_dirs(path, mode=0o755):
    """Creates *path* if it does not exist. Does nothing otherwise."""
    try:
        os.makedirs(path, mode)
    except OSError:
        pass

def discard_file(path):
    """Removes *path* if it exists. Does nothing otherwise."""
    try:
        os.remove(path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise

def iter_file(fp, chunk_size=None):
    """Iterates through a file object in chunks."""
    chunk_size = chunk_size or io.DEFAULT_BUFFER_SIZE
    for chunk in iter(lambda: fp.read(chunk_size), b''):
        yield chunk

def import_package(pkgname):
    """Imports all modules in a package (aka directory).

    Returns a dict of `{<modname>: <exc>}` for modules that raise an exception
    on import.
    """
    exceptions = {}

    path = base_path(pkgname.replace('.', '/'))
    for _, modname, ispkg in pkgutil.iter_modules([path]):
        if ispkg: continue
        exc = import_module(modname, pkgname)
        if exc:
            exceptions[modname] = exc

    return exceptions

def import_module(modname, pkgname=None):
    """Imports a module, optionally from a package.

    Returns any exceptions."""
    name = '.'.join(filter(None, [pkgname, modname]))
    try:
        log.debug('loading: %s', name)
        if pkgname:
            importlib.import_module('.' + modname, pkgname)
        else:
            importlib.import_module(modname)
    except Exception as e:
        return e
